# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
from .callbacks import Callback, Callbacks, Cbs
import k1lib, numpy as np, math
import k1lib.cli as cli
from functools import partial
plt = k1lib.dep("matplotlib.pyplot")
from typing import Callable
__all__ = ["Loss", "Accuracy"]
def plotF(losses, f): # actual function stored by the sliceable plot             # plotF
    plt.figure(figsize=(10, 3), dpi=100); f = f | cli.deref()                    # plotF
    try:                                                                         # plotF
        plt.subplot(1, 2, 1); plt.plot(range(len(losses.train)) | f, losses.train | f); plt.title(f"Train loss") # plotF
        plt.subplot(1, 2, 2); plt.plot(range(len(losses.valid)) | f, losses.valid | f); plt.title(f"Valid loss") # plotF
    except: pass                                                                 # plotF
def commonPlot(obj, f=cli.iden()):                                               # commonPlot
    plotF(obj, f); return                                                        # commonPlot
    return k1lib.viz.SliceablePlot(partial(plotF, obj, f), docs="""\n\nReminder: the actual slice you put in is for the training plot. The valid loss's plot will update automatically to be in the same time frame""") # commonPlot
def nonEmptyList(_list):                                                         # nonEmptyList
    return [0] if _list == [] else _list                                         # nonEmptyList
@k1lib.patch(Cbs)                                                                # nonEmptyList
class Loss(Callback):                                                            # Loss
    " "                                                                          # Loss
    def __init__(self, f=lambda l: l.loss):                                      # Loss
        """Records losses after each batch.
Expected variables in :class:`~k1lib.Learner`:

- loss: single float value

:param f: optional function to get the loss from :class:`~k1lib.Learner` object""" # Loss
        super().__init__(); self.order = 20; self.f = f                          # Loss
        self.train = []; self.valid = [] # all stats all times                   # Loss
        # average stats for each epoch                                           # Loss
        self.epoch = k1lib.Object.fromDict({"train": [], "valid": []})\
                        .withRepr("Use...\n" +\
                                 "- `.train` for epoch-averaged training losses\n" +\
                                 "- `.valid` for epoch-averaged validation losses\n" +\
                                 "- `.plot()` to plot the 2 above")              # Loss
        self.plot = partial(commonPlot, self)                                    # Loss
        self.epoch.plot = partial(commonPlot, self.epoch)                        # Loss
        self._trainLosses = []; self._validLosses = []                           # Loss
        self._landscape = k1lib.callbacks.Landscape(lambda l: l.loss, "_LossLandscape") # Loss
    def endLoss(self):                                                           # Loss
        loss = self.f(self.l)                                                    # Loss
        if self.l.model.training: self._trainLosses.append(loss)                 # Loss
        else: self._validLosses.append(loss)                                     # Loss
    def endEpoch(self):                                                          # Loss
        self.train.extend(self._trainLosses); self.epoch.train.append(np.mean(nonEmptyList(self._trainLosses))) # Loss
        self.valid.extend(self._validLosses); self.epoch.valid.append(np.mean(nonEmptyList(self._validLosses))) # Loss
        self._trainLosses = []; self._validLosses = []                           # Loss
    @property                                                                    # Loss
    def Landscape(self):                                                         # Loss
        """Gets loss-landscape-plotting Callback.
Example::

    l = k1lib.Learner.sample()
    l.cbs.add(Cbs.Loss())
    l.Loss.Landscape.plot()"""                                                   # Loss
        self.cbs.add(self._landscape); return self._landscape                    # Loss
    def detach(self): self._landscape.detach(); return super().detach()          # Loss
    def clear(self):                                                             # Loss
        """Clears saved data"""                                                  # Loss
        self.train = []; self.epoch.train = []                                   # Loss
        self.valid = []; self.epoch.valid = []                                   # Loss
    def __repr__(self):                                                          # Loss
        return f"""{super()._reprHead}, use...
- cb.train: for all training losses over all epochs and batches (#epochs * #batches)
- cb.valid: for all validation losses over all epochs and batches (#epochs * #batches)
- cb.plot(): to plot the 2 above
- cb.clear(): to clear saved data
- cb.epoch: for average losses of each epochs
- cb.Landscape: for loss-landscape-plotting Callback
{super()._reprCan}"""                                                            # Loss
accFMsg = "You have to specify how to compute the accuracy with the AccF callback first" # Loss
@k1lib.patch(Cbs)                                                                # Loss
class Accuracy(Callback):                                                        # Accuracy
    " "                                                                          # Accuracy
    def __init__(self, variable:str="accuracy"):                                 # Accuracy
        """Records accuracies after each batch.
Expected variables in :class:`~k1lib.Learner`:

- accuracy: single float value from 0 to 1

:param variable: name of variable expected to be available in Learner"""         # Accuracy
        super().__init__(); self.order = 20                                      # Accuracy
        self.train = [0]; self.valid = [0]; self.paused = True; self.variable = variable # Accuracy
        self._landscape = k1lib.callbacks.Landscape(lambda l: l.__dict__[variable], "_AccuracyLandscape") # Accuracy
    @property                                                                    # Accuracy
    def hasAccF(self):                                                           # Accuracy
        return any(isinstance(cb, Cbs.AccF) for cb in self.l.cbs.cbs)            # Accuracy
    def startRun(self):                                                          # Accuracy
        self.paused = not self.hasAccF                                           # Accuracy
        if not self.paused:                                                      # Accuracy
            self.train = list(self.train); self.valid = list(self.valid)         # Accuracy
    def endRun(self):                                                            # Accuracy
        if not self.paused:                                                      # Accuracy
            self.train = np.array(self.train); self.valid = np.array(self.valid) # Accuracy
    def endLoss(self):                                                           # Accuracy
        if not self.paused:                                                      # Accuracy
            (self.train if self.l.model.training else self.valid).append(self.l.__dict__[self.variable]) # Accuracy
    def plot(self, f=cli.iden()):                                                # Accuracy
        """
:param f:Optional post-processing cli"""                                         # Accuracy
        if not self.hasAccF: raise RuntimeError(accFMsg)                         # Accuracy
        plt.figure(figsize=(10, 3), dpi=100); f = f | cli.deref()                # Accuracy
        try:                                                                     # Accuracy
            plt.subplot(1, 2, 1); plt.plot(range(len(self.train)) | f, 100*self.train | f); plt.title(f"Train accuracy") # Accuracy
            plt.subplot(1, 2, 2); plt.plot(range(len(self.valid)) | f, 100*self.valid | f); plt.title(f"Valid accuracy") # Accuracy
        except: pass                                                             # Accuracy
    @property                                                                    # Accuracy
    def Landscape(self):                                                         # Accuracy
        """Gets accuracy-landscape-plotting Callback.
Example::

    l = k1lib.Learner.sample()
    l.add(Cbs.Accuracy())
    l.Accuracy.Landscape.plot()

This exact example won't work, as the sample :class:`~k1lib.Learner` task is not
categorical, but the general idea still stands"""                                # Accuracy
        if self.hasAccF:                                                         # Accuracy
            self._landscape.parent = self                                        # Accuracy
            self.cbs.add(self._landscape); return self._landscape                # Accuracy
        else: raise RuntimeError(f"{accFMsg}, before you can view the landscape") # Accuracy
    def clear(self):                                                             # Accuracy
        """Clears saved data."""                                                 # Accuracy
        self.train = [0]; self.valid = [0]                                       # Accuracy
    def __repr__(self):                                                          # Accuracy
        return f"""{super()._reprHead}{f" (.accuracyF not defined yet)" if not self.hasAccF else ""}, use...
- a.train: for train accuracies over all batches
- a.valid: for train accuracies over all batches
- a.plot(): to plot the 2 above
- a.clear(): to clear saved data
- a.Landscape: for loss-landscape-plotting Callback
{super()._reprCan}"""                                                            # Accuracy



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
from .atom import Atom, mo, _a
from functools import partial
from typing import List
import k1lib, re
__all__ = ["alkane", "alcohol", "sameEmpirical"]
# a little bit complex molecules
def _mole(name, f):                                                              # _mole
    _a[name] = f; mo.registerSubstance(name, f)                                  # _mole
_mole("H2O", lambda: mo.O(mo.H)(mo.H))                                           # _mole
_mole("CH4", lambda: mo.C(mo.H)(mo.H)(mo.H)(mo.H))                               # _mole
_mole("COOH", lambda: mo.C(mo.O, 2)(mo.O(mo.H))(mo.H))                           # _mole
_mole("NH3", lambda: mo.N(mo.H)(mo.H)(mo.H))                                     # _mole
_mole("CH3OH", lambda: mo.CH4(mo.H2O))                                           # _mole
def glucose():                                                                   # glucose
    o = mo.O                                                                     # glucose
    o.bond(mo.CH4).bond(mo.CH3OH).bond(mo.CH3OH).bond(mo.CH3OH).bond(mo.CH3OH).bond(o) # glucose
    o.next().bond(mo.CH3OH); return o                                            # glucose
_mole("glucose", glucose)                                                        # glucose
def cyclohexane():                                                               # cyclohexane
    c = mo.CH4                                                                   # cyclohexane
    return c.bond(mo.CH4).bond(mo.CH4).bond(mo.CH4).bond(mo.CH4).bond(mo.CH4).bond(c) # cyclohexane
_mole("cyclohexane", cyclohexane)                                                # cyclohexane
def benzene():                                                                   # benzene
    c = mo.CH4                                                                   # benzene
    return c.main(mo.CH4, 2).main(mo.CH4).main(mo.CH4, 2).main(mo.CH4).main(mo.CH4, 2).main(c) # benzene
_mole("benzene", benzene)                                                        # benzene
def adenine():                                                                   # adenine
    n = mo.NH3                                                                   # adenine
    c1 = n.bond(mo.CH4).bond(mo.NH3, 2).bond(mo.CH4)                             # adenine
    c2 = c1.bond(mo.CH4, 2); c2.bond(n)                                          # adenine
    c1.bond(mo.CH4)(mo.NH3).bond(mo.NH3, 2).bond(mo.CH4).bond(mo.NH3, 2).bond(c2); # adenine
    return n                                                                     # adenine
_mole("adenine", adenine)                                                        # adenine
def ribose():                                                                    # ribose
    o = mo.H2O                                                                   # ribose
    return o.bond(mo.CH4(mo.CH3OH)).main(mo.CH3OH).main(mo.CH3OH).main(mo.CH3OH).main(o) # ribose
_mole("ribose", ribose)                                                          # ribose
def adenosine(): ri = mo.ribose; ri.next(1)(mo.adenine); return ri               # adenosine
_mole("adenosine", adenosine)                                                    # adenosine
def alkane(n:int) -> Atom:                                                       # alkane
    """Creates an alkane with ``n`` carbons.
Example::

    # returns "C3H8"
    mo.alkane(3).empirical()"""                                                  # alkane
    answer = mo.CH4                                                              # alkane
    for i in range(n-1): answer = answer.bond(mo.CH4)                            # alkane
    return answer                                                                # alkane
def alcohol(n:int, loc:int=1) -> Atom:                                           # alcohol
    """Creates an alcohol with ``n`` carbons and an OH group at carbon ``loc``.
Example::

    # returns "C3H8O"
    mo.alcohol(3, 1).empirical()"""                                              # alcohol
    a = alkane(n)                                                                # alcohol
    a.next(times=loc-1)(mo.H2O)                                                  # alcohol
    return a                                                                     # alcohol
@k1lib.patch(Atom)                                                               # alcohol
def perfluoro_ize(mainA:Atom) -> Atom:                                           # perfluoro_ize
    """Replaces all C-H bonds with C-F bonds"""                                  # perfluoro_ize
    for a in mainA.atoms():                                                      # perfluoro_ize
        b = a.next()                                                             # perfluoro_ize
        if a.name == "H" and b.name == "C":                                      # perfluoro_ize
            a.removeBond(b); b(mo.F)                                             # perfluoro_ize
    return mainA                                                                 # perfluoro_ize
empC = re.compile("[A-Z][a-z]*[0-9]*")                                           # perfluoro_ize
def sameEmpirical(a:str, b:str) -> bool:                                         # sameEmpirical
    """Checks whether 2 empirical formula are the same or not.
Example::

    moparse.sameEmpirical("C2H4", "H4C2") # returns True"""                      # sameEmpirical
    return sorted(re.findall(empC, a)) == sorted(re.findall(empC, b))            # sameEmpirical



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""For not very complicated accuracies functions"""
from ..callbacks import Callback, Callbacks, Cbs
from typing import Callable, Tuple
import k1lib
try: import torch; hasTorch = True
except: torch = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {})); hasTorch = False
__all__ = ["AccF"]
AccFSig = Callable[[Tuple[torch.Tensor, torch.Tensor]], float]
PredFSig = Callable[[torch.Tensor], torch.Tensor]
@k1lib.patch(Cbs)
class AccF(Callback):                                                            # AccF
    " "                                                                          # AccF
    def __init__(self, predF:PredFSig=None, accF:AccFSig=None, integrations:bool=True, variable:str="accuracy", hookToLearner:bool=True): # AccF
        """Generic accuracy function.

Built in default accuracies functions are fine, if you don't do something too
dramatic/different. Expected variables in :class:`~k1lib.Learner`:

- y:  :class:`~torch.Tensor` of shape (\*N, C)
- yb: :class:`~torch.Tensor` of shape (\*N,)

Deposits variables into :class:`~k1lib.Learner`:

- preds:      detached, batched tensor output of ``predF``
- accuracies: detached, batched tensor output of ``accF``
- accuracy:   detached, single float, mean of ``accuracies``

Where:

- N is the batch size. Can be multidimensional, but has to agree between ``y`` and ``yb``
- C is the number of categories

:param predF: takes in ``y``, returns predictions (tensor with int elements indicating the categories)
:param accF: takes in ``(predictions, yb)``, returns accuracies (tensor with 0 or 1 elements)
:param integrations: whether to integrate :class:`~k1lib.callbacks.confusionMatrix.ConfusionMatrix` or not.
:param variable: variable to deposit into Learner"""                             # AccF
        super().__init__(); self.order = 10; self.integrations = integrations; self.ownsConMat = False; self.hookToLearner = hookToLearner # AccF
        self.predF = predF or (lambda y: y.argmax(-1))                           # AccF
        self.accF = accF or (lambda p, yb: (p == yb)+0); self.variable = variable # AccF
    def attached(self):                                                          # AccF
        if self.integrations:                                                    # AccF
            if "ConfusionMatrix" not in self.cbs:                                # AccF
                self.conMatCb = Cbs.ConfusionMatrix()                            # AccF
                self.cbs.add(self.conMatCb); self.ownsConMat = True              # AccF
            else: self.conMatCb = self.cbs.ConfusionMatrix                       # AccF
    def endLoss(self):                                                           # AccF
        preds = self.predF(self.l.y); accs = self.accF(preds, self.l.yb);        # AccF
        if self.hookToLearner:                                                   # AccF
            self.l.preds = preds.detach()                                        # AccF
            self.l.accuracies = accs.detach()                                    # AccF
        self.l.__dict__[self.variable] = accs.float().mean().item()              # AccF
    def detach(self):                                                            # AccF
        super().detach()                                                         # AccF
        if self.conMatCb != None:                                                # AccF
            if self.ownsConMat: self.conMatCb.detach()                           # AccF
            self.conMatCb = None                                                 # AccF



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""
This module is for color formats, units and whatnot. This is exposed
automatically with::

   from k1lib.imports import *
   fmt.txt # exposed
"""
import k1lib, math, re; from k1lib import cli
from typing import Dict, Iterator, Tuple
pygments = k1lib.dep("pygments")
__all__ = ["generic", "metricPrefixes", "size", "fromSize", "sizeOf",
           "comp", "compRate", "time", "item", "throughput", "txt", "code", "h", "pre", "row", "col", "colors", "rmAnsi"]
k1lib.settings.add("fmt", k1lib.Settings().add("separator", True, "whether to have a space between the number and the unit"), "from k1lib.fmt module");
settings = k1lib.settings.fmt
metricPrefixes = {-8:"y",-7:"z",-6:"a",-5:"f",-4:"p",-3:"n",-2:"u",-1:"m",0:"",1:"k",2:"M",3:"G",4:"T",5:"P",6:"E",7:"Z",8:"Y"}
#metricPrefixes = ["", "k", "M", "G", "T", "P", "E", "Z", "Y"]
def generic(x, units:Dict[int, str]):                                            # generic
    c = " " if settings.separator else ""                                        # generic
    for i, unit in units.items():                                                # generic
        upperBound = 1000 * 1000**i                                              # generic
        if abs(x) < upperBound:                                                  # generic
            return f"{round(1e3*x/upperBound, 2)}{c}{unit}"                      # generic
    return (f"{round(1e3*x/upperBound, 2)}{c}{unit}").strip()                    # generic
sizes = {i: f"{p}B" for i, p in metricPrefixes.items() if i >= 0}; #sizes[0] = "bytes" # generic
def size(_bytes=0):                                                              # size
    """Formats disk size.
Example::

    # returns "50.0 bytes"
    fmt.size(50)
    # returns "12.0 MB"
    fmt.size(1.2e7)
"""                                                                              # size
    return generic(_bytes, sizes)                                                # size
sizeInv = {"k": 1e3, "m": 1e6, "g": 1e9, "t": 1e12, "p": 1e15, "e": 1e18, "z": 1e21, "y": 1e24} # size
def fromSize(s:str) -> int: # this is ugly I know, doesn't fit well into others. But making a generalized version seems hard, and I just need this right now # fromSize
    """Grabs size from string representation.
Example::

    fromSize("31.5k") # returns 31500
    fromSize("31.5kB") # also returns 31500
"""                                                                              # fromSize
    s = s.lower().replace(" ", "").rstrip("b"); ch = s[-1]                       # fromSize
    if ch in sizeInv: return int(float(s[:-1])*sizeInv[ch])                      # fromSize
    return int(s)                                                                # fromSize
def sizeOf(l:Iterator[float]) -> Tuple[str, Iterator[float]]:                    # sizeOf
    """Figures out appropriate scale, scales back the Iterator, and return both.
Example::

    x = torch.abs(torch.randn(2)) * 1e4 + 1e5
    label, t = fmt.sizeOf(x) # label is "kB"
    (t | toTensor()).min() # min value should be close to 100"""                 # sizeOf
    l = list(l | cli.apply(lambda n: abs(n)))                                    # sizeOf
    v = l | cli.toMax()                                                          # sizeOf
    v = math.log10(v) if v > 0 else -math.log10(-v)                              # sizeOf
    idx = math.floor(v/3)                                                        # sizeOf
    coef = 1.0/1000**idx                                                         # sizeOf
    return sizes[idx], l | cli.apply(lambda x: x * coef) | cli.deref()           # sizeOf
computations = {i: f"{p}FLOPs" for i, p in metricPrefixes.items() if i >= 0}     # sizeOf
def comp(flop=0):                                                                # comp
    """Formats computation amount.
Example::

    # returns "50.0 FLOPs"
    fmt.computation(50)
    # returns "50.0 MFLOPs"
    fmt.computation(5e7)
"""                                                                              # comp
    return generic(flop, computations)                                           # comp
computationRates = {i: f"{p}FLOPS" for i, p in metricPrefixes.items() if i >= 0} # comp
def compRate(flops=0):                                                           # compRate
    """Formats computation rate.
Example::

    # returns "50.0 FLOPS"
    fmt.computationRate(50)
    # returns "50.0 MFLOPS"
    fmt.computationRate(5e7)
"""                                                                              # compRate
    return generic(flops, computationRates)                                      # compRate
times = {i:f"{p}s" for i, p in metricPrefixes.items() if i <= 0}                 # compRate
def time(seconds=0):                                                             # time
    """Formats small times.
Example::

    fmt.time(50) # returns "50.0 s"
    fmt.time(4000) # returns "4000.0 s"
    fmt.time(0.02) # returns "20.0 ms"
    fmt.time(1e-5) # returns "10.0 us"
"""                                                                              # time
    return generic(seconds, times)                                               # time
items = {0: "", 1: "k", 2: "M", 3: "B", 4: "T"}                                  # time
def item(n=0):                                                                   # item
    """Formats generic item.
Example::

    # returns "50.0"
    fmt.item(50)
    # returns "500.0 k"
    fmt.item(5e5)
"""                                                                              # item
    return generic(n, items)                                                     # item
def throughput(n, unit=""):                                                      # throughput
    """Formats item throughput.
Example::

    # returns "3.16/year"
    fmt.throughput(1e-7)
    # returns "2.63/month"
    fmt.throughput(1e-6)
    # returns "3.6/hour"
    fmt.throughput(1e-3)
    # returns "100.0 k/s"
    throughput(1e5)

    # returns "100.0 k epochs/s"
    throughput(1e5, " epochs")

:param n: items per second
:param unit: optional item unit"""                                               # throughput
    if n < 10/(365.25*86400): return item(n*(365.25*86400)) + f"{unit}/year"     # throughput
    if n < 10/(30.4375*86400): return item(n*(30.4375*86400)) + f"{unit}/month"  # throughput
    if n < 10/86400: return item(n*86400) + f"{unit}/day"                        # throughput
    if n < 10/3600: return item(n*3600) + f"{unit}/hour"                         # throughput
    if n < 10/60: return item(n*60) + f"{unit}/minute"                           # throughput
    return item(n) + f"{unit}/s"                                                 # throughput
_esc = '\033['                                                                   # throughput
_end = f'{_esc}0m'                                                               # throughput
class txt:                                                                       # txt
    """Text formatting.
Example::

    # will print out red text
    print(fmt.txt.red("some text"))"""                                           # txt
    @staticmethod                                                                # txt
    def darkcyan(s:str):  return f"{_esc}36m{s}{_end}"                           # txt
    @staticmethod                                                                # txt
    def red(s:str):       return f"{_esc}91m{s}{_end}"                           # txt
    @staticmethod                                                                # txt
    def green(s:str):     return f"{_esc}92m{s}{_end}"                           # txt
    @staticmethod                                                                # txt
    def yellow(s:str):    return f"{_esc}93m{s}{_end}"                           # txt
    @staticmethod                                                                # txt
    def blue(s:str):      return f"{_esc}94m{s}{_end}"                           # txt
    @staticmethod                                                                # txt
    def purple(s:str):    return f"{_esc}95m{s}{_end}"                           # txt
    @staticmethod                                                                # txt
    def cyan(s:str):      return f"{_esc}96m{s}{_end}"                           # txt
    @staticmethod                                                                # txt
    def bold(s:str):      return f"{_esc}1m{s}{_end}"                            # txt
    @staticmethod                                                                # txt
    def grey(s:str):      return f"{_esc}38;2;150;150;150m{s}{_end}"             # txt
    @staticmethod                                                                # txt
    def darkgrey(s:str):  return f"{_esc}38;2;100;100;100m{s}{_end}"             # txt
    @staticmethod                                                                # txt
    def underline(s:str): return f"{_esc}4m{s}{_end}"                            # txt
    @staticmethod                                                                # txt
    def identity(s:str):  return f"{s}"                                          # txt
class code:                                                                      # code
    def python(code:str):                                                        # code
        """Formats Python code.
Example::

    fmt.code.python(\"\"\"
    def f(x:int):
        return x + 3
    \"\"\") | aS(IPython.display.HTML)
"""                                                                              # code
        a = "<style>" + pygments.formatters.HtmlFormatter().get_style_defs(".highlight") + "</style>" # code
        b = pygments.highlight(code, pygments.lexers.PythonLexer(), pygments.formatters.HtmlFormatter()) # code
        return a + b                                                             # code
def h(code:str, level:int=1) -> str:                                             # h
    """Wraps content inside a 'h' html tag.
Example::

    fmt.h("abc", 2) # returns "<h2>abc</h2>"

:param level: what's the header level?"""                                        # h
    return f"<h{level}>{code}</h{level}>"                                        # h
def _jsF_h(meta, level=3):                                                       # _jsF_h
    fIdx = cli.init._jsFAuto(); dataIdx = cli.init._jsDAuto()                    # _jsF_h
    return f"const {fIdx} = ({dataIdx}) => `<h{level}>${{{dataIdx}}}</h{level}>`", fIdx # _jsF_h
k1lib.settings.cli.kjs.jsF[h] = _jsF_h                                           # _jsF_h
def pre(code:str, extras:str="") -> str:                                         # pre
    """Wraps content inside a 'pre' html tag.
Example::

    fmt.pre("abc")
"""                                                                              # pre
    return f"<pre style='font-family: monospace' {extras} >{code}</pre>"         # pre
def _jsF_pre(meta):                                                              # _jsF_pre
    fIdx = cli.init._jsFAuto(); dataIdx = cli.init._jsDAuto()                    # _jsF_pre
    return f"const {fIdx} = ({dataIdx}) => `<pre style='font-family: monospace'>${{{dataIdx}}}</pre>`", fIdx # _jsF_pre
k1lib.settings.cli.kjs.jsF[pre] = _jsF_pre                                       # _jsF_pre
def col(args, margin=10):                                                        # col
    """Creates a html col of all the elements.
Example::

    fmt.col(["abc", "def"]) | aS(IPython.display.HTML)
"""                                                                              # col
    return args | cli.apply(lambda x: f"<div style='margin: {margin}px'>{x}</div>") | cli.join("") | cli.aS(lambda x: f"<div style='display: flex; flex-direction: column'>{x}</div>") # col
def _jsF_col(meta, margin=10):                                                   # _jsF_col
    fIdx = cli.init._jsFAuto(); dataIdx = cli.init._jsDAuto()                    # _jsF_col
    fIdx2 = cli.init._jsFAuto(); dataIdx2 = cli.init._jsDAuto()                  # _jsF_col
    return f"""
const {fIdx2} = ({dataIdx2}) => `<div style='margin: {margin}px'>${{{dataIdx2}}}</div>`
const {fIdx}  = ({dataIdx})  => `<div style='display: flex; flex-direction: column'>${{{dataIdx}.map({fIdx2}).join('')}}</div>`
    """, fIdx                                                                    # _jsF_col
k1lib.settings.cli.kjs.jsF[col] = _jsF_col                                       # _jsF_col
def row(args, margin=10):                                                        # row
    """Creates a html row of all the elements.
Example::

    fmt.row(["abc", "def"]) | aS(IPython.display.HTML)
"""                                                                              # row
    return args | cli.apply(lambda x: f"<div style='margin: {margin}px'>{x}</div>") | cli.join("") | cli.aS(lambda x: f"<div style='display: flex; flex-direction: row'>{x}</div>") # row
def _jsF_row(meta, margin=10):                                                   # _jsF_row
    fIdx = cli.init._jsFAuto(); dataIdx = cli.init._jsDAuto()                    # _jsF_row
    fIdx2 = cli.init._jsFAuto(); dataIdx2 = cli.init._jsDAuto()                  # _jsF_row
    return f"""
const {fIdx2} = ({dataIdx2}) => `<div style='margin: {margin}px'>${{{dataIdx2}}}</div>`
const {fIdx}  = ({dataIdx})  => `<div style='display: flex; flex-direction: row'>${{{dataIdx}.map({fIdx2}).join('')}}</div>`
    """, fIdx                                                                    # _jsF_row
k1lib.settings.cli.kjs.jsF[row] = _jsF_row                                       # _jsF_row
settings.add("colors", ["#8dd3c7", "#ffffb3", "#bebada", "#fb8072", "#80b1d3", "#fdb462", "#b3de69", "#fccde5", "#d9d9d9", "#bc80bd", "#ccebc5", "#ffed6f"], "List of colors to cycle through in fmt.colors()") # _jsF_row
def colors():                                                                    # colors
    """Returns an infinite iterator that cycles through 12 colors.
Example::

    fmt.colors() | head(3) | deref()

Color scheme taken from https://colorbrewer2.org/#type=qualitative&scheme=Set3&n=12""" # colors
    return settings.colors | cli.repeatFrom()                                    # colors
ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')               # colors
def rmAnsi(text):                                                                # rmAnsi
    """Removes ansi escape characters, courtesy of https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python.""" # rmAnsi
    return ansi_escape.sub('', text)                                             # rmAnsi



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""
This module is for Lua's serpent module, which serializes object into a string
similar to json. But maddingly, no one has actually wrote a serialization/deserialization
code in Python before, and I desperately need it for a factorio project. Because
this was written in a rush, I'm in no way guaranteeing that it will work on all
serialized objects out there, but from my testing, it seems pretty robust. This
is exposed automatically with::

   from k1lib.imports import *
   serpent.loads(...) # exposed
"""
import k1lib, json, re, io; from collections import deque
__all__ = ["deconstruct", "listCorrection", "loads_monolith", "loads_fragments", "loads", "dumps"]
def _deconstruct(s:str, i:int, meta, stack, ctx, autoInc):                       # _deconstruct
    x = s.find("{", i); y = s.find("}", i)                                       # _deconstruct
    if y < 0: return -1, None                                                    # _deconstruct
    if x < y and x >= 0: # opens a new context and saves parent context to stack # _deconstruct
        if ctx[3] is None: ctx[3] = "=" in s[ctx[1]:x] # not yet determined whether it's a list or dict # _deconstruct
        stack.append(ctx); ctx = [autoInc(), x, None, None, ctx[0]]; return x+1, ctx # _deconstruct
    else: # closes the current context and go back to the parent context         # _deconstruct
        if ctx[3] is None: ctx[3] = "=" in s[ctx[1]:y]                           # _deconstruct
        ctx[2] = y; meta.append(ctx); ctx = stack.pop(); return y+1, ctx         # _deconstruct
def deconstruct(s:str) -> "list[list[5]]":                                       # deconstruct
    """Not intended for the end user. Deconstructs and grab metadata of some
lua objects. Example::

    a = '{1,2,3,{a=3,b={c=6,d={5,6,7}}},{b=3}}'
    serpent.deconstruct(a)

That returns::

    [['_v3', 21, 27, False, '_v2'],
     ['_v2', 14, 28, True, '_v1'],
     ['_v1', 7, 29, True, '_v0'],
     ['_v4', 31, 35, True, '_v0'],
     ['_v0', 0, 36, False, 'root']]

The columns are: [unique index of bracket, start byte, end byte, is it a dictionary?, parent index]

This is a crucial step within :meth:`listCorrection`"""                          # deconstruct
    autoInc = k1lib.AutoIncrement(prefix="_v"); meta = []; stack = deque()       # deconstruct
    i = 0; n = len(s); ctx = ["root", None, None, None, None] # [idx, sB, eB, isDict?] # deconstruct
    while i < n: i, ctx = _deconstruct(s, i, meta, stack, ctx, autoInc)          # deconstruct
    return meta                                                                  # deconstruct
def listCorrection(s:str) -> str:                                                # listCorrection
    """Not intended for the end user. Corrects for lists in Lua.
Example::

    a = '{1,2,3,{a=3,b={c=6,d={5,6,7}}},{b=3}}'
    serpent.listCorrection(a) # returns '[1,2,3,{a=3,b={c=6,d=[5,6,7]}},{b=3}]'

See how some pointy brackets have been replaced with square brackets?

This is because there are no list or tuple types in lua, and there are also no sets
in json, so kinda have to roll my own solution"""                                # listCorrection
    b = bytearray(s.encode())                                                    # listCorrection
    for idx, sB, eB, isDict, pIdx in deconstruct(s):                             # listCorrection
        if not isDict: b[sB] = b"["[0]; b[eB] = b"]"[0]                          # listCorrection
    return bytes(b).decode()                                                     # listCorrection
# def _loads(lua):                                                               # listCorrection
#     matches = re.findall("[a-zA-Z_\-\"\[\]]+[ ]*=", lua); replacements = [m.rstrip("= ").strip("[]\"'") for m in matches] # listCorrection
#     for x, y in zip(matches, replacements): lua = lua.replace(x, f'"{y}": ')   # listCorrection
#     return json.loads(lua.replace(" ", ""))                                    # listCorrection
p1 = re.compile("[a-zA-Z0-9_\-\"\[\]]+[ ]*=")                                    # listCorrection
def loads_monolith(lua:str) -> object:                                           # loads_monolith
    """Not intended for the end user. Core loading mechanism. See :meth:`loads`""" # loads_monolith
    lua = listCorrection(lua); matches = sorted(re.findall(p1, lua), key=lambda x: -len(x)) # loads_monolith
    replacements = [m.rstrip("= ").strip("[]\"'") for m in matches]              # loads_monolith
    for x, y in zip(matches, replacements): lua = lua.replace(x, f'"{y}": ')     # loads_monolith
    return json.loads(lua)                                                       # loads_monolith
def loads_fragments(lua:str) -> object:                                          # loads_fragments
    """Not intended for the end user. See :meth:`loads`. Deserializes lua objects,
breaking up the work into multiple fragments. So here's the general gist::

    s = "{1, 2, 3, {4, 5, 6}, {7, 8, 9}}"

    # then, we grab the fragments, which are the top level {} blocks, assigning a unique key (the character \ue000 and autoInc index)
    fragments = {"\ue0000": {4, 5, 6}, "\ue0001": {7, 8, 9}}
    # then, we replace the fragments with their keys
    s = "{1, 2, 3, "\ue0000", "\ue0001"}"

    # then we load s, it will run fast since the fragments are just simple strings
    s = serpent.loads_monolith(s)

    # then we patch s, replacing the keys with actual parsed objects
    s = {1, 2, 3, {4, 5, 6}, {7, 8, 9}}

Why so convoluted? Well turns out, loads_monolith is pretty slow. It has a for loop there,
and there's a .replace() within, which is a hidden for loop that copies the entire
string over and over again, which slows it down. Haven't done extensive testing, but
feels like O(n^2) time complexity while I was working with it.

So this optimization assumes that the top level {} blocks are small, but there're many
of them, thus this assigns less work (shorter string, hence faster .replace()) to each
loads_monolith() calls. So if there're 10k fragments, this can potentially be 10k faster.

This assumption of course is not that great and not very general, and you can easily find
ways around it. But it's just enough for my use case right now, which is to analyze factorio.
The correct way would be to dive deeper and benchmark everything more clearly, but I don't
have time for that."""                                                           # loads_fragments
    # splits into fragments                                                      # loads_fragments
    a = [[f"\ue000{i}",loads_monolith(lua[sB:eB+1]), sB, eB] for i,[idx,sB,eB,d,pIdx] in enumerate(deconstruct(lua)) if pIdx == "_v0"] # loads_fragments
                                                                                 # loads_fragments
    # fast way of replacing the fragments with the pua unicodes                  # loads_fragments
    s = io.StringIO(); lastSeek = 0; d = {pua:o for pua, o, sB, eB in a}         # loads_fragments
    for pua,obj,sB,eB in a: s.write(lua[lastSeek:sB]); s.write(f'"{pua}"'); lastSeek = eB+1 # loads_fragments
    s.write(lua[lastSeek:]); s.seek(0); c = loads_monolith(s.read())             # loads_fragments
                                                                                 # loads_fragments
    # injecting                                                                  # loads_fragments
    if isinstance(c, dict): return {k:d.get(v, v) for k,v in c.items()}          # loads_fragments
    elif isinstance(c, list): return [d.get(v, v) for v in c]                    # loads_fragments
def preprocess(lua:str): return "".join([x.split("--")[0] for x in lua.split("\n")]).replace(" ", "").replace("\\", "\\\\") # preprocess
def loads(lua:str):                                                              # loads
    """Deserialize lua objects from string.
Example::

    # returns [1, 2, 3, {'a': 3, 'b': {'c': 6, 'd': [5, 6, 7]}}, {'b': 3}]
    loads("{ 1, 2, 3, { a = 3, b = { c = 6, d = {5, 6, 7} } }, { b = 3 } }")

See also: :meth:`dumps`

What's the relative speed here? Because everything is written in Python, I
expect it to be slower than json, but by how much? Here're some benchmark results:

|             | lua  | json | binary |
| ----------- | ---- | ---- | ------ |
| from python | 21us | 11us | 184us  |
| to   python | 92us | 10us | 5.8us  |

The "lua" column uses "serpent.loads()", "json" uses "json.loads()", and
"binary" uses "dill.loads()"
"""                                                                              # loads
    lua = preprocess(lua)                                                        # loads
    return loads_fragments(lua) if len(lua) > 1000 else loads_monolith(lua)      # loads
def _dumps_dict(obj):                                                            # _dumps_dict
    for k,v in obj.items(): k = f'["{k}"]' if "-" in k else k; yield f"{k}={dumps(v)}" # _dumps_dict
def dumps(obj) -> str:                                                           # dumps
    """Serialize Python objects into lua string.
Example::

    # returns '{1,2,3,{a=3,b={c=6,d={5,6,7}}},{b=3}}'
    serpent.dumps([1, 2, 3, {'a': 3, 'b': {'c': 6, 'd': [5, 6, 7]}}, {'b': 3}])

See also: :meth:`loads`"""                                                       # dumps
    if isinstance(obj, dict): return f"{{{','.join(_dumps_dict(obj))}}}"         # dumps
    if isinstance(obj, (list, tuple)): return f"{{{','.join([dumps(e) for e in obj])}}}" # dumps
    return json.dumps(obj)                                                       # dumps



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""This is for everything related to ipython notebooks. Expected to use behind the
"nb" module name, like this::

    from k1lib.imports import *
    nb.execute("file.ipynb")"""
__all__ = ["cells", "grabTags", "executeTags", "pretty", "execute"]
from k1lib.cli import BaseCli; import k1lib.cli as cli
import json, k1lib, os, traceback; from typing import List
from collections import defaultdict
plt = k1lib.dep("matplotlib.pyplot")
def _cells(fileName, outputs=False):                                             # _cells
    js = json.loads(cli.cat(fileName) | cli.join("\n"))                          # _cells
    cells = []; fields = set(["cell_type", "source"])                            # _cells
    if outputs: fields.add("outputs")                                            # _cells
    return [{k:cell[k] for k in cell.keys() if k in fields} for cell in js["cells"]] # _cells
def cells(fileName=None, outputs=False):                                         # cells
    """Gets simplified notebook cells from file source, including fields
``cell_type`` and ``source`` only. Example::

    nb.cells("file.ipynb")"""                                                    # cells
    if fileName is None: return cli.aS(_cells, outputs)                          # cells
    else: return _cells(fileName, outputs)                                       # cells
def _tagHandler(*a,**kw): pass                                                   # _tagHandler
def grabTags(s:str):                                                             # grabTags
    """Gets the tags included in a cell.
Example::

    nb.grabTags("# serve(node='mint-6.l'), abc") # returns ["serve", "abc"]
"""                                                                              # grabTags
    t = defaultdict(lambda: _tagHandler)                                         # grabTags
    s = s.strip("#").strip()                                                     # grabTags
    try: eval(f'[{s}]', t)                                                       # grabTags
    except: return [e.split("(")[0].strip() for e in s.split(",")]               # grabTags
    return [e for e in t.keys() if not (e.startswith("__") and e.endswith("__"))] # grabTags
def executeTags(s:str, funcs:dict):                                              # executeTags
    """Executes the tags included in a cell.
Example::

    d = []
    def serve(node): d.append(node)
    nb.executeTags("# serve(node='mint-6.l'), abc", {"serve": serve, "abc": 123})

    d # now is ["mint-6.l"]
"""                                                                              # executeTags
    try: s = s.strip("#").strip(); eval(f'[{s}]', dict(funcs))                   # executeTags
    except: pass                                                                 # executeTags
class pretty(BaseCli):                                                           # pretty
    def __init__(self, magics:bool=False, whitelist:List[str]=[], blacklist:List[str]=[]): # pretty
        """Makes the cells prettier.
Cell 1 in file.ipynb::

    #notest, export
    a = 3

Cell 2 in file.ipynb::

    b = 6

Code::

    # only cell 2 gets chosen
    nb.cells("file.ipynb") | nb.pretty(blacklist=["notest"])
    # only cell 1 gets chosen
    nb.cells("file.ipynb") | nb.pretty(whitelist=["export"])

:param magics: if False, then if detected magics ('!', '%%' symbols), then remove that
    line in cell's source
:param whitelist: every cell that doesn't have any of these properties will be
    filtered out
:param blacklist: every cell that has any of these properties will be filtered out""" # pretty
        self.magics = magics; self.wl = whitelist; self.bl = blacklist           # pretty
    def __ror__(self, cells):                                                    # pretty
        magics = self.magics; wl = self.wl; bl = self.bl                         # pretty
        for cell in cells:                                                       # pretty
            lines = cell["source"]                                               # pretty
            if not magics: lines = [line.rstrip() for line in lines if not line.startswith("!") and not line.startswith("%%")] # pretty
            if len(lines) == 0: continue                                         # pretty
            if lines[0].startswith("#"): props = set(grabTags(lines[0]))         # pretty
            else: props = set()                                                  # pretty
            if len(wl) > 0 and not any([e in props for e in wl]): continue       # pretty
            if len(bl) > 0 and any([e in props for e in bl]): continue           # pretty
            cell["source"] = lines; yield cell                                   # pretty
class execute(BaseCli):                                                          # execute
    def __init__(self, fileName=None, _globals:dict=None):                       # execute
        """Executes cells.
Example::

    nb.cells("file.ipynb") | nb.execute("nb.ipynb")

Most of the time, you'd want to pass cells through :class:`pretty` first, to make sure
everything is nice and clean

:param fileName: not actually used to read the file. If specified, then changes the
    current working directory to that of the file
:param _globals: optional dict of global variables"""                            # execute
        self.fileName = fileName                                                 # execute
        self._globals = dict() if _globals is None else _globals                 # execute
    def __ror__(self, cells):                                                    # execute
        with k1lib.settings.context():                                           # execute
            if self.fileName:                                                    # execute
                k1lib.settings.wd = os.path.dirname(self.fileName) or "."        # execute
            for cell in cells:                                                   # execute
                if cell["cell_type"] != "code": continue                         # execute
                source = "\n".join(cell["source"] | cli.apply(lambda x: x.strip("\n"))) # execute
                try: exec(source, self._globals); plt.show()                     # execute
                except Exception as e:                                           # execute
                    print("Problematic cell:\n")                                 # execute
                    print(source); traceback.print_exc(); raise e                # execute
    @staticmethod                                                                # execute
    def rightAway(fileName:str, _globals:dict=None, tag:str="export"):           # execute
        """Convenience function to execute a notebook right away.
Example::

    fn = "some/file.ipynb"
    nb.cells(fn) | nb.pretty(whitelist=[tag]) | nb.execute(nb, globals())
    nb.execute.rightAway(fn, globals(), tag)

Last 2 lines are pretty much the same."""                                        # execute
        if tag: tag = pretty(whitelist=[tag]) if tag else cli.iden()             # execute
        return cells(fileName) | tag | execute(fileName, _globals)               # execute



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
import k1lib as k1, os
import k1lib.cli as cli
from collections import deque
from functools import lru_cache
from contextlib import contextmanager
__all__ = ["sql", "s3", "s3bucket"]
@contextmanager
def mysqlCnf(user, password, host, port):                                        # mysqlCnf
    fn = f"""[client]\nuser = "{user}"\npassword = "{password or ''}"\nhost = "{host}"\nport = "{port}" """ | cli.file() # mysqlCnf
    try: yield fn                                                                # mysqlCnf
    finally: os.remove(fn)                                                       # mysqlCnf
mysqlConn = k1.dep("mysql.connector", "mysql-connector-python", "https://pypi.org/project/mysql-connector-python/") # mysqlCnf
pgConn = k1.dep("psycopg2", "psycopg2-binary", "https://pypi.org/project/psycopg2/") # mysqlCnf
sqlite3 = k1.dep("sqlite3")                                                      # mysqlCnf
qD = {"my": "`", "pg": "", "lite": ""} # quote dict                              # mysqlCnf
class sql:                                                                       # sql
    def __init__(self, host="127.0.0.1", port=3306, user=None, password="", mode="my"): # sql
        """Creates a connection to a SQL database.
Example::

    s = sql("127.0.0.1") # creates a new sql object. Apparently, mysql and mysqldump on the command line bugs out if you put "localhost". If you put localhost here, it should work fine, but `sql(...) | toBytes()` and other functions that use external programs can deny you access
    s.refresh()          # refreshes connection any time you encounter strange bugs

    s | ls()                                                       # returns List[sqldb], lists out all databases
    s | toBytes()                                                  # returns Iterator[str], dumps every databases. Yes, it's "toBytes", but it has the feeling of serializing whatever the input is, so it's at least intuitive in that way
    "dump.sql" | s                                                 # restores the database using the dump file
    cat("dump.sql") | s                                            # restores the database using the dump file
    db1 = s["db1"]                                                 # returns sqldb, gets database named "db1"
    db1 | ls()                                                     # returns List[sqltable], list out all tables within this database
    db1 | toBytes()                                                # returns Iterator[str], dumps the database
    users = db1["user"]                                            # gets table named "user", short and simple

    db1.query("select * from users")                     # queries the database using your custom query
    db1.query("select * from users where user_id=%s", 3) # queries with prepared statement

    users.info()                                               # prints out the first 10 rows of the table and the table schema
    users.cols                                                 # returns table's columns as List[str]
    len(users)                                                 # returns number of rows
    users.query(...)                                           # can also do a custom query, just like with databases
    users[1]                                                   # grabs user with id 1. Returns None if user doesn't exist
    users[1:10]                                                # grabs users with id from 1 (inclusive) to 10 (exclusive)
    users[1].firstname = "Reimu"                               # sets the first name of user with id 1 to "Reimu"
    users[1] = {"firstname": "Reimu", "lastname": "Hakurei"}   # sets first name and last name of user with id 1
    users | toBytes()                                          # dumps this specific table, returns Iterator[str]
    users |  cat() | display()                                 # reads entire table, gets first 10 rows and displays it out
    users | (cat() | head(20)) | display()                     # reads first 20 rows only, then displays the first 10 rows. Query sent is "select * from user limit 20"
    users | (cat() | filt("x == 4", 3) | head(20)) | display() # grabs first 20 rows that has the 4th column equal to 4, then displays the first 10 rows. Query sent is "select user_id, address, balance, age from user where age = 4", assuming the table only has those columns

Philosophy for these methods is that they should be intuitive to look at and interact with, not for performance.
If performance is needed, just write raw queries or don't use a database and use applyCl() instead.

:param host: host name, ip address, or file name (in case of sqlite)
:param port: port at the host
:param uesr: database user name. If not specified then fallback to environment variable ``SQL_USER``, then ``USER``
:param password: database password. If not specified then assume database doesn't require one
:param mode: currently supports 3 values: "my" (MySQL), "pg" (PostgreSQL) and "lite" (SQLite)
"""                                                                              # sql
        self.host = host; self.port = port; self.user = user or os.environ.get("SQL_USER") or os.environ.get("USER") # sql
        self.password = password or os.environ.get("SQL_PASSWORD"); self.db = None; self.mode = mode; self.conn = None; self.refresh() # sql
    def refresh(self):                                                           # sql
        try: self.conn.close(); self.conn = None                                 # sql
        except: pass                                                             # sql
        kwargs = dict(host=self.host, port=self.port, user=self.user, password=self.password, database=self.db) # sql
        if self.mode == "my": self.conn = mysqlConn.connect(**kwargs, charset='utf8mb4', collation='utf8mb4_general_ci') # sql
        elif self.mode == "pg": self.conn = pgConn.connect(**kwargs)             # sql
        elif self.mode == "lite": self.conn = sqlite3.connect(self.host)         # sql
        else: raise Exception(f"Can only support 'my' (MySQL), 'pg' (PostgreSQL) and 'lite' (SQLite) for now, can't support {self.mode}") # sql
    def _changeDb(self, db):                                                     # sql
        if self.db != db:                                                        # sql
            if self.mode == "my": self.query(f"use `{db}`")                      # sql
            elif self.mode == "pg": self.db = db; self.refresh()                 # sql
            elif self.mode == "lite": pass                                       # sql
        self.db = db                                                             # sql
    def query(self, query, *args):                                               # sql
        cur = self.conn.cursor(); cur.execute(query, args)                       # sql
        try: ans = cur.fetchall()                                                # sql
        except: ans = None                                                       # sql
        cur.close(); self.conn.commit(); return ans                              # sql
    def _ls(self):                                                               # sql
        if self.mode == "my": return [sqldb(self, e[0]) for e in self.query("show databases")] # sql
        elif self.mode == "pg": return [sqldb(self, e[0]) for e in self.query("select datname from pg_database where datistemplate=false")] # sql
        elif self.mode == "lite": return [sqldb(self, "default")]                # sql
    def __repr__(self): return f"<sql mode={self.mode} host={self._host}>"       # sql
    @property                                                                    # sql
    def _host(self) -> str:                                                      # sql
        if self.mode == "lite": return self.host.split(os.sep)[-1]               # sql
        else: return f"{self.host}:{self.port}"                                  # sql
    def _cnfCtx(self): return mysqlCnf(self.user, self.password, self.host, self.port) # sql
    def _toBytes(self):                                                          # sql
        if self.mode == "my":                                                    # sql
            with self._cnfCtx() as fn: yield from None | cli.cmd(f"mysqldump --defaults-file={fn} --single-transaction --hex-blob --all-databases") # sql
        else: raise Exception(f"All databases dump of mode {self.mode} is not supported yet") # sql
    def __ror__(self, it): # restoring a backup                                  # sql
        if self.mode == "my":                                                    # sql
            def restore(fn):                                                     # sql
                with self._cnfCtx() as cnfFn: None | cli.cmd(f"mysql --defaults-file={cnfFn} < {fn}") | cli.ignore() # sql
            if isinstance(it, str):     restore(it)                              # sql
            else: fn = it | cli.file(); restore(fn); os.remove(fn)               # sql
        else: raise Exception(f"Restoring database from .sql file of mode {self.mode} is not supported yet") # sql
    def __len__(self): return len(self._ls())                                    # sql
    def __getitem__(self, idx):                                                  # sql
        if isinstance(idx, str):                                                 # sql
            lsres = [x for x in self._ls() if x.name == idx]                     # sql
            if len(lsres) == 1: return lsres[0]                                  # sql
            else: raise Exception(f"Database '{idx}' does not exist, can't retrieve it") # sql
        return self._ls()[idx]                                                   # sql
    def __iter__(self): return iter(self._ls())                                  # sql
    def __delitem__(self, idx):                                                  # sql
        if isinstance(idx, str):                                                 # sql
            lsres = [x for x in self._ls() if x.name == idx]                     # sql
            if len(lsres) == 1:                                                  # sql
                if self.mode == "my" or self.mode == "pg": return self.query(f"drop database {idx}") # sql
                else: raise Exception("Currently only support dropping databases in mysql and postgresql") # sql
            else: raise Exception(f"Database '{idx}' does not exist, can't delete it") # sql
        else: raise Exception("Only supports deleting with table names. Use this like `del db['some_table']`") # sql
class sqldb:                                                                     # sqldb
    def __init__(self, sql:sql, name:str):                                       # sqldb
        """A sql database representation. Not expected to be instatiated by you. See also: :class:`sql`""" # sqldb
        self.sql = sql; self.name = name                                         # sqldb
    def query(self, query, *args): self.sql._changeDb(self.name); return self.sql.query(query, *args) # sqldb
    def _ls(self):                                                               # sqldb
        if self.sql.mode == "my": return [sqltable(self.sql, self, e[0]) for e in self.query(f"show tables")] # sqldb
        if self.sql.mode == "pg": return [sqltable(self.sql, self, e[0]) for e in self.query(f"select table_name from information_schema.tables")] # sqldb
        if self.sql.mode == "lite": return [sqltable(self.sql, self, e[0]) for e in self.query("select name from sqlite_master where type='table'")] # sqldb
    def __repr__(self): return f"<sqldb host={self.sql._host} db={self.name}>"   # sqldb
    def _toBytes(self):                                                          # sqldb
        if self.sql.mode == "my":                                                # sqldb
            with self.sql._cnfCtx() as fn: yield from None | cli.cmd(f"mysqldump --defaults-file={fn} --single-transaction --hex-blob --databases {self.name}") # sqldb
        else: raise Exception(f"Database dump of mode {self.sql.mode} is not supported yet") # sqldb
    def __ror__(self, it): return self.sql.__ror__(it)                           # sqldb
    def __len__(self): return len(self._ls())                                    # sqldb
    def __getitem__(self, idx):                                                  # sqldb
        if isinstance(idx, str):                                                 # sqldb
            lsres = [x for x in self._ls() if x.name == idx]                     # sqldb
            if len(lsres) == 1: return lsres[0]                                  # sqldb
            else: raise Exception(f"Table '{idx}' does not exist, can't retrieve it") # sqldb
        return self._ls()[idx]                                                   # sqldb
    def __iter__(self): return iter(self._ls())                                  # sqldb
    def __delitem__(self, idx):                                                  # sqldb
        if isinstance(idx, str):                                                 # sqldb
            lsres = [x for x in self._ls() if x.name == idx]                     # sqldb
            if len(lsres) == 1: return self.query(f"drop table {idx}")           # sqldb
            else: raise Exception(f"Table '{idx}' does not exist, can't delete it") # sqldb
        else: raise Exception("Only supports deleting with table names. Use this like `del db['some_table']`") # sqldb
class sqltable:                                                                  # sqltable
    def __init__(self, sql, sqldb, name:str):                                    # sqltable
        """A sql table representation. Not expected to be instantiated by you. See also: :class:`sql`""" # sqltable
        self.sql = sql; self.sqldb = sqldb; self.name = name; self._cols = None  # sqltable
    def _cat(self, ser):                                                         # sqltable
        cols = self.cols; _2 = [] # clis that can't be optimized, stashed away to be merged with ser later on # sqltable
        q = qD[self.sql.mode]; clis = deque(ser.clis)                            # sqltable
        o1 = None # cut() opt                                                    # sqltable
        o2 = None # head() opt                                                   # sqltable
        o3 = [] # filt() opt                                                     # sqltable
        while len(clis) > 0:                                                     # sqltable
            c = clis.popleft()                                                   # sqltable
            if isinstance(c, cli.filt): _2.append(c); break # TODO: add optimizations for filt # sqltable
            elif o2 is None and isinstance(c, cli.head):                         # sqltable
                if round(c.n) != c.n or c.n < 0 or c.inverted or c.n == None: _2.append(c); break # sqltable
                else: o2 = f"limit {c.n}"; continue                              # sqltable
            elif o1 is None and isinstance(c, cli.cut):                          # sqltable
                if isinstance(c.columns, slice): _2.append(c); o1 = 0; continue  # sqltable
                else:                                                            # sqltable
                    o1 = ", ".join([f"{q}{c}{q}" for c in cols | cli.rows(*c.columns)]) # sqltable
                    if len(c.columns) == 1: _2.append(cli.item().all() | cli.aS(list)) # sqltable
            else: _2.append(c); break                                            # sqltable
        o1 = o1 or ", ".join([f"{q}{c}{q}" for c in cols])                       # sqltable
        query = f"select {o1} from {q}{self.name}{q} {o2 or ''}"#; print(f"query: {query}"); return [] # sqltable
        sql = self.sql; return [sqlrow(sql, self, row) for row in self.sqldb.query(query) | cli.serial(*_2, *clis)] # sqltable
    @property                                                                    # sqltable
    def cols(self):                                                              # sqltable
        """Get column names"""                                                   # sqltable
        if not self._cols: self._cols = self.describe()[1:] | cli.cut({"my": 0, "pg": 0, "lite": 1}[self.sql.mode]) | cli.deref() # sqltable
        return self._cols                                                        # sqltable
    @lru_cache                                                                   # sqltable
    def describe(self):                                                          # sqltable
        if self.sql.mode == "my": return self.sqldb.query(f"describe `{self.name}`") | cli.insert(["Field", "Type", "Null", "Key", "Default", "Extra"]) | cli.deref() # sqltable
        if self.sql.mode == "pg": return self.sqldb.query(f"select column_name, data_type, is_nullable, column_default, ordinal_position from information_schema.columns where table_name='{self.name}'") | cli.insert(["column_name", "data_type", "is_nullable", "column_default", "ordinal_position"]) | cli.deref() # sqltable
        if self.sql.mode == "lite": return self.sqldb.query(f"pragma table_info([{self.name}])") | cli.insert(["cid", "name", "type", "notnull", "dflt_value", "pk"]) | cli.deref() # sqltable
    def insert(self, **kwargs):                                                  # sqltable
        """Inserts a row.
Example::

    table = ...
    table.insert(firstname="Yuyuko", lastname="Saigyouji")
"""                                                                              # sqltable
        keys = ", ".join([f"`{x}`" for x in kwargs.keys()])                      # sqltable
        values = ", ".join(['%s']*len(kwargs))                                   # sqltable
        self.query(f"insert into {self.name} ({keys}) values ({values})", *kwargs.values()) # sqltable
    def _update(self, idx, **kwargs):                                            # sqltable
        """Updates a row.
Example::

    rable = ...
    table.update(3, firstname="Youmu", lastname="Konpaku")
"""                                                                              # sqltable
        p1 = ", ".join([f"`{k}` = %s" for k in kwargs.keys()])                   # sqltable
        self.query(f"update {self.name} set {p1} where {self.cols[0]} = {idx}", *kwargs.values()) # sqltable
    def info(self, out=False):                                                   # sqltable
        """Preview table.
Example::

    table = ...
    table.info()

:param out: if True, returns a list of lines instead of printing them out"""     # sqltable
        def gen():                                                               # sqltable
            print(f"Table `{self.name}`\n")                                      # sqltable
            desc = self.describe() | cli.deref(); cols = self.cols; q = qD[self.sql.mode]; s = ", ".join([f"{q}{e}{q}" for e in cols]) # sqltable
            self.sqldb.query(f"select {s} from {q}{self.name}{q} limit 9") | (cli.aS(repr) | cli.head(50)).all(2) | cli.insert(cols) | cli.display(); print("") # sqltable
            desc | cli.display(None)                                             # sqltable
        if out:                                                                  # sqltable
            with k1.captureStdout() as out: gen()                                # sqltable
            return out()                                                         # sqltable
        else: gen()                                                              # sqltable
    def query(self, query, *args): return self.sqldb.query(query, *args)         # sqltable
    def __repr__(self): return f"<sqltable host={self.sql._host} db={self.sqldb.name} table={self.name}>" # sqltable
    def _toBytes(self):                                                          # sqltable
        if self.sql.mode == "my":                                                # sqltable
            with self.sql._cnfCtx() as fn: yield from None | cli.cmd(f"mysqldump --defaults-file={fn} --single-transaction --hex-blob {self.sqldb.name} {self.name}") # sqltable
        else: raise Exception(f"Table dump of mode {self.sql.mode} is not supported yet") # sqltable
    def __ror__(self, it): return self.sql.__ror__(it)                           # sqltable
    def __len__(self): return self.query(f"select count(*) from {self.name}")[0][0] # sqltable
    def __getitem__(self, idx):                                                  # sqltable
        idName = self.cols[0]                                                    # sqltable
        if isinstance(idx, int):                                                 # sqltable
            res = self.query(f"select * from {self.name} where {idName} = {idx}") # sqltable
            if len(res) == 0: return None                                        # sqltable
            return sqlrow(self.sql, self, res[0])                                # sqltable
        elif isinstance(idx, slice):                                             # sqltable
            idxs = list(range(idx.stop if idx.stop is not None else (self.query(f"select max({idName}) from {self.name}")[0][0]+1))[idx]) # sqltable
            res = self.query(f"select * from {self.name} where {idName} in ({', '.join([str(x) for x in idxs])})") # sqltable
            d = {row[0]:row for row in res}; cols = self.cols; return [sqlrow(self.sql, self, d.get(idx, None)) for idx in idxs] # sqltable
        else: raise Exception("Only support table indexing of integers or slices") # sqltable
    def __setitem__(self, idx, value):                                           # sqltable
        if not isinstance(value, dict): raise Exception("Only accepts setting elements with dicts") # sqltable
        if self[idx] is None: raise Exception(f"Can't set element with id {idx}, it doesn't seem to exist. Use table.insert(col1=value1, col2=value2) instead") # sqltable
        self._update(idx, **value)                                               # sqltable
    def __iter__(self): return iter(self | cli.cat())                            # sqltable
    def __delitem__(self, idx):                                                  # sqltable
        idName = self.cols[0]                                                    # sqltable
        if isinstance(idx, int):                                                 # sqltable
            res = self[idx]                                                      # sqltable
            if res: self.query(f"delete from {self.name} where {idName} = %s", res[0]) # sqltable
        elif isinstance(idx, slice):                                             # sqltable
            res = self[idx]; ids = ", ".join([row[0] for row in res if row])     # sqltable
            self.query(f"delete from {self.name} where {idName} in {ids}")       # sqltable
        else: raise Exception("Only support table indexing of integers or slices") # sqltable
class sqlrow:                                                                    # sqlrow
    def __init__(self, sql, sqltable, row):                                      # sqlrow
        self._ab_sentinel = True                                                 # sqlrow
        self.sql = sql; self.sqltable = sqltable                                 # sqlrow
        self.row = row; self._d = None                                           # sqlrow
        self._ab_sentinel = False                                                # sqlrow
    def __getitem__(self, idx): return self.row[idx]                             # sqlrow
    def __getattr__(self, attr):                                                 # sqlrow
        if self._d is None: self._d = {c:r for c,r in zip(self.sqltable.cols, self.row)} # sqlrow
        return self._d[attr]                                                     # sqlrow
    def __setattr__(self, attr, value):                                          # sqlrow
        if attr == "_ab_sentinel": self.__dict__[attr] = value                   # sqlrow
        else:                                                                    # sqlrow
            if self._ab_sentinel or attr in self.__dict__: self.__dict__[attr] = value # sqlrow
            else: return self.sqltable._update(self.row[0], **{attr: value})     # sqlrow
    def __len__(self): return len(self.row)                                      # sqlrow
    def __repr__(self): return f"{self.row}"                                     # sqlrow
boto3 = k1.dep("boto3")                                                          # sqlrow
class s3:                                                                        # s3
    def __init__(self, client):                                                  # s3
        """Represents an S3 client.
Example::

    client = boto3.client("s3", ...)            # put your credentials and details here
    db = s3(client)                             # creates an S3 manager
    db | ls()                                   # lists all buckets accessible

    bucket = db | ls() | item()                 # grabs the first bucket, returns object of type s3bucket
    bucket = s3bucket(client, "bucket-name")    # or you can instantiate the bucket directly
    bucket | ls()                               # lists all objects within this bucket
    bucket | ls() | grep("\\.so")               # grabs all .so files from the bucket

    obj = bucket | ls() | item()                # grabs the first object within this bucket, returns object of type s3obj
    obj.key, obj.size, obj.lastModified         # some fields directly accessible

This mostly offers interoperability with ls() and cat(), so that you can
write relatively intuitive code, but fundamentally provides no upsides"""        # s3
        self.client = client                                                     # s3
    def _ls(self): return [s3bucket(self.client, x["Name"]) for x in self.client.list_buckets()["Buckets"]] # s3
    def __repr__(self): return f"<kaws.s3 client>"                               # s3
class s3bucket:                                                                  # s3bucket
    def __init__(self, client, name:str):                                        # s3bucket
        """Represents an S3 bucket.
See also: :class:`s3`"""                                                         # s3bucket
        self.client = client; self.name = name                                   # s3bucket
    def _ls(self): client = self.client; name = self.name; return [s3obj(client, name, data) for data in self.client.list_objects(Bucket=name).get("Contents", [])] # s3bucket
    def __repr__(self): return f"<s3bucket name='{self.name}'>"                  # s3bucket
class s3obj:                                                                     # s3obj
    def __init__(self, client, bucket:str, data):                                # s3obj
        """Represents an S3 object. Not intended to be instantiated directly.
See also: :class:`s3`"""                                                         # s3obj
        self.client = client; self.bucket = bucket                               # s3obj
        self.key  = data["Key"];  self.lastModified = data["LastModified"]       # s3obj
        self.size = data["Size"]; self.storageClass = data["StorageClass"]       # s3obj
    def __repr__(self): return f"<s3obj bucket='{self.bucket}' key='{self.key}' size={fmt.size(self.size)}>" # s3obj
    def _cat(self, kwargs):                                                      # s3obj
        if kwargs["text"] is not False: raise Exception(f"s3obj does not support `cat(text=True)`. Please use `cat(text=False)` instead") # s3obj
        if kwargs["chunks"]: raise Exception(f"s3obj does not support `cat(chunks=True)`") # s3obj
        sB = kwargs["sB"]; eB = kwargs["eB"]                                     # s3obj
        if eB < 0: eB = self.size                                                # s3obj
        return self.client.get_object(Bucket=self.bucket, Key=self.key, Range=f'bytes={sB}-{eB-1}')["Body"].read() # s3obj



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
__all__ = ["grep", "grepTemplate"]
import re, k1lib, json
from k1lib.cli.init import BaseCli; import k1lib.cli.init as init
import k1lib.cli.kjs as kjs; import k1lib.cli as cli
from collections import deque; from typing import Iterator, Union, Callable, Any
def extGuard(cond, s=""):                                                        # extGuard
    if cond: Exception(f"Can't use extract mode. {s}")                           # extGuard
def jsfGuard(cond, s=""):                                                        # jsfGuard
    if cond: Exception(f"grep._jsF() does not support custom {s} parameter")     # jsfGuard
inf = float("inf")                                                               # jsfGuard
class grep(BaseCli):                                                             # grep
    def __init__(self, pattern:Union[str, Callable[[Any], bool]], before:int=0, after:int=0, N:int=float("inf"), sep:bool=False, col:int=None, extract:str=None): # grep
        """Find lines that has the specified pattern.
Example::

    # returns ['d', 'd']
    "abcde12d34" | grep("d") | deref()
    # returns ['c', 'd', '2', 'd'], 2 sections of ['c', 'd'] and ['2', 'd']
    "abcde12d34" | grep("d", 1) | deref()
    # returns ['c', 'd']
    "abcde12d34" | grep("d", 1, N=1) | deref()
    # returns ['d', 'e', 'd', '3', '4'], 2 sections of ['d', 'e'] and ['d', '3', '4']
    "abcde12d34" | grep("d", 0, 3).till("e") | deref()
    # returns [['0', '1', '2'], ['3', '1', '4']]
    "0123145" | grep("1", 2, 1, sep=True) | deref()

You can also separate out the sections::

    # returns [['c', 'd'], ['2', 'd']]
    "abcde12d34" | grep("d", 1, sep=True) | deref()
    # returns [['c', 'd']]
    "abcde12d34" | grep("d", 1, N=1, sep=True) | deref()
    # returns [['1', '2', '3'], ['1', '4', '5']]
    "0123145" | grep("1", sep=True).till() | deref()

You can also put in predicates instead of regex patterns::

    # returns ['d', 'd']
    "abcde12d34" | grep(lambda x: x == "d") | deref()
    # also returns ['d', 'd']
    "abcde12d34" | filt(lambda x: x == "d") | deref()
    # returns ['d', 'e', 'd', '3', '4']
    "abcde12d34" | grep(lambda x: x == "d").till(lambda x: x == "e") | deref()

The first scenario looks like a regular filter function, already implemented by :class:`~k1lib.cli.filt.filt`,
but :class:`grep` brings in more clustering features for the price of reduced
execution speed. So for simple scenarios it's advised that you use :class:`~k1lib.cli.filt.filt`.

See also: :class:`~k1lib.cli.structural.groupBy`

Also, there's a `whole tutorial <../tutorials/cli.html>`_ devoted to just this cli

Also also, if each element in the input iterator is not a string/bytes, and
you're searching using regex, then it will get its representation and searches
in it.

.. admonition:: Extract mode

    Sometimes, you want to extract a subsection of a matched string, like extracting
    links in a html file, then you can do something like this::

        # returns ['a.io', 'b.com', 'c.net']
        ["href='a.io'", "href='b.com'", "href='c.net'"] | grep("href='(?P<g>.*)'", extract="g") | deref()
        # returns [['a.io', 3], ['b.com', 4], ['c.net', 5]]
        [["href='a.io'", 3], ["href='b.com'", 4], ["href='c.net'", 5]] | grep("href='(?P<g>.*)'", extract="g", col=0) | deref()

    Essentially, you're defining the group with name "g" to be any string within a quote block
    following "href", and then it will just extract out the group that you want. Because the
    purpose of this mode is to extract matched objects, a few of the arguments don't really make
    sense and thus are disabled, like "before", "after", "sep", "N"

Regex quick cheatsheet:

- `\\d`: digit (\\D for inverse)
- `^`: begin of string ($ for end of string)
- `\\w`: unicode word (\\W for inverse)
- `(?!...)`: matches if the inside does not match
- `(?P<name>...)`: matches as group "name"
- `A|B`: matches A or B
- `[aml]`: set of characters "a", "m" and "k"
- `a{3,5}`: matches character "a" 3 to 5 times ("aaa", "aaaa" and "aaaaa")
- `a*`: matches "a" 0 or more times (`a*?` matches "a" 0 or more times non-greedy)

:param pattern: regex pattern to search for in a line
:param before: lines before the hit. Outputs independent lines
:param after: lines after the hit. Outputs independent lines
:param N: max sections to output
:param sep: whether to separate out the sections as lists
:param col: searches for pattern in a specific column"""                         # grep
        super().__init__(); self.pattern = pattern                               # grep
        if isinstance(pattern, str):                                             # grep
            self._f = re.compile(pattern).search; self.mode = 0 # make func quickly accessible # grep
        else: self._f = cli.op.solidify(pattern); self.mode = 1 # mode for either regex or normal funcs # grep
        self.before = before; self.after = after; self.col = col; self.N = N; self.sep = sep; self.inverted = False # grep
        self.tillPattern = None; self.tillAfter = None; self._tillF = lambda x: False; self.extract = extract # grep
        if extract:                                                              # grep
            extGuard(before, "`before` has to be zero")                          # grep
            extGuard(after, "`after` has to be zero")                            # grep
            extGuard(sep, "`sep` has to be False")                               # grep
            extGuard(col is not None, "`col` has to be None. Cut out the column if you want") # grep
            extGuard(N < inf, "`N` has to be infinite. Just use head() if you want to limit the number of results") # grep
    def till(self, pattern:Union[str, Callable[[Any], bool]]=None):              # grep
        """Greps until some other pattern appear. Inclusive, so you might want to
trim the last line. Example::

    # returns ['5', '6', '7', '8'], includes last item
    range(10) | join("") | grep("5").till("8") | deref()
    # returns ['d', 'e', 'd', '3', '4']
    "abcde12d34" | grep("d").till("e") | deref()
    # returns ['d', 'e']
    "abcde12d34" | grep("d", N=1).till("e") | deref()

If initial pattern and till pattern are the same, then you don't have use this method at
all. Instead, do something like this::

    # returns ['1', '2', '3']
    "0123145" | grep("1", after=1e9, N=1) | deref()"""                           # grep
        if self.extract: extGuard(True, "Can't use .till() in extract mode as it makes no sense") # grep
        if pattern is None: self._tillF = self._f                                # grep
        elif isinstance(pattern, str): self._tillF = re.compile(pattern).search  # grep
        else: self._tillF = cli.op.solidify(pattern)                             # grep
        self.tillAfter = self.after; self.after = inf; return self               # grep
    def __ror__(self, it:Iterator[str]) -> Iterator[str]:                        # grep
        self.sectionIdx = 0; col = self.col; _f = self._f; _tillF = self._tillF  # grep
        if self.sep:                                                             # grep
            elems = []; idx = 0                                                  # grep
            s = self._clone(); s.sep = False                                     # grep
            for line in (it | s):                                                # grep
                if s.sectionIdx > idx: # outputs whatever remaining              # grep
                    if len(elems) > 0: yield list(elems)                         # grep
                    idx = s.sectionIdx; elems = []                               # grep
                elems.append(line)                                               # grep
            yield list(elems); return                                            # grep
        if self.extract:                                                         # grep
            group = self.extract                                                 # grep
            if col is None:                                                      # grep
                for line in it:                                                  # grep
                    res = _f(line)                                               # grep
                    if res: yield res.group(group)                               # grep
            else:                                                                # grep
                for line in it:                                                  # grep
                    line = list(line); res = _f(line[col])                       # grep
                    if res: line[col] = res.group(group); yield line             # grep
                                                                                 # grep
            return                                                               # grep
        queue = deque([], self.before); counter = 0 # remaining lines after to display # grep
        cRO = k1lib.RunOnce(); cRO.done()                                        # grep
        for line in it:                                                          # grep
            if col != None: line = list(line); elem = line[col]                  # grep
            else: elem = line                                                    # grep
            if self.mode == 0 and not isinstance(elem, (str, bytes)): elem = f"{elem}" # grep
            if _f(elem): # new section                                           # grep
                self.sectionIdx += 1; counter = self.after+1; cRO.revert()       # grep
                if self.sectionIdx > self.N: return                              # grep
                yield from queue; queue.clear(); yield line                      # grep
            elif _tillF(elem) and counter == inf: # closing section              # grep
                counter = self.tillAfter + 1; cRO.revert(); yield line           # grep
            if counter == 0:                                                     # grep
                queue.append(line) # saves recent past lines                     # grep
            elif counter > 0: # yielding "after" section                         # grep
                if cRO.done(): yield line                                        # grep
                counter -= 1                                                     # grep
    def __invert__(self):                                                        # grep
        """Flips the pattern, just like how :class:`~k1lib.cli.filt.filt`
works. Example::

    # returns ['a', 'b', 'c', 'e', '1', '2', '3', '4']
    "abcde12d34" | ~grep("d") | deref()"""                                       # grep
        if self.extract: extGuard(True, "Can't invert search condition in extract mode as it makes no sense") # grep
        if self.inverted: raise Exception("Can't invert grep() more than twice, it makes no sense") # grep
        ans = self._clone(); f = self._f; ans._f = lambda s: not f(s)            # grep
        ans.preInvGrep = self; ans.inverted = True; return ans                   # grep
    def _clone(self):                                                            # grep
        answer = grep(self.pattern, self.before, self.after, self.N, self.sep, self.col) # grep
        answer._tillF = self._tillF; answer.tillAfter = self.tillAfter; return answer # grep
    def _jsF(self, meta):                                                        # grep
        jsfGuard(self.before != 0, "before");  jsfGuard(self.after != 0, "after") # grep
        jsfGuard(self.N < float("inf"), "N");  jsfGuard(self.sep, "sep")         # grep
        jsfGuard(not self.extract is None, "extract"); inverted = False          # grep
        if self.inverted: inverted = True; self = self.preInvGrep                # grep
        if not isinstance(self.pattern, str): raise Exception(f"grep._jsF() does not support pattern that's not a string") # grep
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto(); argIdx = init._jsDAuto(); p = self.pattern # grep
        for x,y in self.pattern | grep("\ue157", sep=True).till("\ue239") | cli.apply(cli.join("")) | cli.filt("x") | cli.apply(lambda x: [x, x.replace("\ue157", "${").replace("\ue239", "}")]): p = p.replace(x, y) # grep
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.grep(`{p}`, {{col: {cli.kjs.v(self.col)}, inv: {json.dumps(inverted)}}})", fIdx # grep
class grepTemplate(BaseCli):                                                     # grepTemplate
    def __init__(self, pattern:str, template:str):                               # grepTemplate
        """Searches over all lines, pick out the match, and expands
it to the templateand yields"""                                                  # grepTemplate
        super().__init__()                                                       # grepTemplate
        self.pattern = re.compile(pattern); self.template = template             # grepTemplate
    def __ror__(self, it:Iterator[str]):                                         # grepTemplate
        super().__ror__(it)                                                      # grepTemplate
        for line in it:                                                          # grepTemplate
            matchObj = self.pattern.search(line)                                 # grepTemplate
            if matchObj is None: continue                                        # grepTemplate
            yield matchObj.expand(self.template)                                 # grepTemplate



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""This is for optimizing the hell out of cli tools. Optimizations
that focus around a specific cli should be included close to their
definitions, so this is for optimizations that unusually span multiple
clis, and serve as examples of how to create optimization passes.

See over the `LLVM optimizer tutorial <llvm.html>`_ for more background."""
import k1lib.cli as cli
from k1lib.cli.typehint import *
from k1lib.cli.init import yieldT
from collections import defaultdict
import k1lib, platform

__all__ = ["dummy"]
def dummy():                                                                     # dummy
    """Does nothing. Only here so that you can read the source code"""           # dummy
#tOpt.clearPasses(); tOpt.n = 10                                                 # dummy
def setHints(cs, ts, metadata):                                                  # setHints
    s = cs[0]; t = ts[0]; #print(f"hints {metadata} - {[c.__class__.__name__ for c in s.clis]}") # setHints
    for c in s.clis: c._inHint = t; t = c._typehint(t) or tAny(); c._outHint = t # setHints
tOpt.addPass(setHints, [cli.serial], 14) # for adding extra hints to cli objects # setHints
def oUnwrapSerial(cs, ts, metadata):                                             # oUnwrapSerial
    if len(metadata["route"]) < 2: return None                                   # oUnwrapSerial
    s = metadata["route"][-2]                                                    # oUnwrapSerial
    if s == "apply" or s == "mtmS" or s == "oneToMany": return None              # oUnwrapSerial
    #print(f"unwrap serial {metadata} {[c.__class__.__name__ for c in cs[0].clis]}") # oUnwrapSerial
    return cs[0].clis                                                            # oUnwrapSerial
tOpt.addPass(oUnwrapSerial , [cli.serial], 1) # for unwrapping serial            # oUnwrapSerial
tOpt.addPass(lambda cs, ts, _: [cs[1], cs[0]], [cli.toList, cli.head]) # for swapping heads around # oUnwrapSerial
def stripSerial(c):                                                              # stripSerial
    if c is None: return None                                                    # stripSerial
    while isinstance(c, cli.serial) and len(c.clis) == 1: c = c.clis[0]          # stripSerial
    return c                                                                     # stripSerial
def prepareSerial(c): # basically prepares input clis so that they can be recursively optimized by `apply` and `mtmS` # prepareSerial
    c = stripSerial(c); return cli.serial(c) if isinstance(c, cli.serial) else cli.serial(cli.serial(c)) # prepareSerial
def oApply(cs, ts, metadata): # for going into apply                             # oApply
    a = cs[0]; t = ts[0]; #print(f"apply {metadata} {cs} {ts} {a.f.clis}")       # oApply
    if a.column is None and isinstance(a.f, cli.BaseCli):                        # oApply
        metadata["route"].append("apply");                                       # oApply
        res = stripSerial(cli.typehint.serialOpt(prepareSerial(a.f), t.item(), metadata)) # oApply
        metadata["route"].pop(); #print(f"res: {res}")                           # oApply
        if res is not None: return [cli.apply(res)]                              # oApply
    return None                                                                  # oApply
tOpt.addPass(oApply, [cli.apply])                                                # oApply
def oMtmS(cs, ts, metadata):                                                     # oMtmS
    m = cs[0]; n = len(m.clis); newClis = []; atLeastOnce = False                # oMtmS
    ts = m._inpTypeHintExpand(ts[0])                                             # oMtmS
    metadata["route"].append("mtmS")                                             # oMtmS
    for c, t in zip(m.clis, ts):                                                 # oMtmS
        res = stripSerial(cli.typehint.serialOpt(prepareSerial(c), t, metadata)) # oMtmS
        if res is not None: atLeastOnce = True                                   # oMtmS
        newClis.append(res)                                                      # oMtmS
    metadata["route"].pop()                                                      # oMtmS
    if atLeastOnce: return [cli.mtmS(*newClis)]                                  # oMtmS
tOpt.addPass(oMtmS, [cli.mtmS])                                                  # oMtmS
def oOneToMany(cs, ts, metadata):                                                # oOneToMany
    o = cs[0]; t = ts[0]; atLeastOnce = False; newClis = []                      # oOneToMany
    metadata["route"].append("oneToMany")                                        # oOneToMany
    for c in o.clis:                                                             # oOneToMany
        res = stripSerial(cli.typehint.serialOpt(prepareSerial(c), t, metadata)) # oOneToMany
        if res is not None: atLeastOnce = True                                   # oOneToMany
        newClis.append(res)                                                      # oOneToMany
    metadata["route"].pop()                                                      # oOneToMany
    if atLeastOnce: return [cli.oneToMany(*newClis)]                             # oOneToMany
tOpt.addPass(oOneToMany, [cli.oneToMany])                                        # oOneToMany
def basics():                                                                    # basics
    tOpt.addPass(oApply, [cli.apply])                                            # basics
    tOpt.addPass(oMtmS, [cli.mtmS])                                              # basics
    tOpt.addPass(oOneToMany, [cli.oneToMany])                                    # basics
def oFileLength(cs, ts, _):                                                      # oFileLength
    c, s = cs;                                                                   # oFileLength
    if s.idx != 0: return None                                                   # oFileLength
    return [cli.aS(lambda fn: None | cli.cmd(f"wc -l {fn}") | cli.item() | cli.op().split(" ")[0].ab_int())] # oFileLength
if platform.system() in ["Linux", "Darwin"]: tOpt.addPass(oFileLength, [cli.cat().__class__, cli.shape], 9) # oFileLength



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
from typing import List, Iterator, Any, NewType, TypeVar, Generic
import k1lib.cli as cli; from numbers import Number
import k1lib, itertools, copy, xml, warnings, traceback, sys, random; import numpy as np
from collections import deque
import xml.etree.ElementTree
try: import torch; hasTorch = True
except: hasTorch = False; torch = k1lib.dep("torch")

__all__ = ["BaseCli", "Table", "T", "fastF", "yieldT",
           "serial", "oneToMany", "mtmS"]
settings = k1lib.Settings()
atomic = k1lib.Settings()
settings.add("atomic", atomic, "classes/types that are considered atomic and specified cli tools should never try to iterate over them")
settings.add("defaultDelim", "\t", "default delimiter used in-between columns when creating tables. Defaulted to tab character.")
settings.add("defaultIndent", "  ", "default indent used for displaying nested structures")
settings.add("strict", False, "turning it on can help you debug stuff, but could also be a pain to work with")
settings.add("inf", float("inf"), "infinity definition for many clis. Here because you might want to temporarily not loop things infinitely")
k1lib.settings.add("cli", settings, "from k1lib.cli module")
yieldT = object()
def patchDefaultDelim(st:str):                                                   # patchDefaultDelim
    """
:param s:
    - if not None, returns self
    - else returns the default delimiter in :attr:`~k1lib.settings`"""           # patchDefaultDelim
    return settings.defaultDelim if st is None else st                           # patchDefaultDelim
def patchDefaultIndent(st:str):                                                  # patchDefaultIndent
    """
:param s:
    - if not None, returns self
    - else returns the default indent character in :attr:`~k1lib.settings`"""    # patchDefaultIndent
    return settings.defaultIndent if st is None else st                          # patchDefaultIndent
T = TypeVar("T")                                                                 # patchDefaultIndent
"""Generic type variable"""                                                      # patchDefaultIndent
class _MetaType(type):                                                           # _MetaType
    def __getitem__(self, generic):                                              # _MetaType
        d = {"__args__": generic, "_n": self._n, "__doc__": self.__doc__}        # _MetaType
        return _MetaType(self._n, (), d)                                         # _MetaType
    def __repr__(self):                                                          # _MetaType
        def main(self):                                                          # _MetaType
            def trueName(o):                                                     # _MetaType
                if isinstance(o, _MetaType): return main(o)                      # _MetaType
                try: return o.__name__                                           # _MetaType
                except: return f"{o}"                                            # _MetaType
            if hasattr(self, "__args__"):                                        # _MetaType
                if isinstance(self.__args__, tuple):                             # _MetaType
                    return f"{self._n}[{', '.join([trueName(e) for e in self.__args__])}]" # _MetaType
                else: return f"{self._n}[{trueName(self.__args__)}]"             # _MetaType
            return self._n                                                       # _MetaType
        return main(self)                                                        # _MetaType
def newTypeHint(name, docs=""):                                                  # newTypeHint
    """Creates a new type hint that can be sliced and yet still looks fine
in sphinx. Crudely written by my poorly understood idea of Python's
metaclasses. Seriously, this shit is bonkers, read over it https://stackoverflow.com/questions/100003/what-are-metaclasses-in-python

Example::

    Table = newTypeHint("Table", "some docs")
    Table[int] # prints out as "Table[int]", and sphinx fell for it too
    Table[Table[str], float] # prints out as "Table[Table[str], float]"
"""                                                                              # newTypeHint
    return _MetaType(name, (), {"_n": name, "__doc__": docs})                    # newTypeHint
#Table = newTypeHint("Table", """Essentially just Iterator[List[T]]. This class is just here so that I can generate the docs with nicely formatted types like "Table[str]".""") # newTypeHint
#Table = NewType("Table", List)                                                  # newTypeHint
class Table(Generic[T]):                                                         # Table
    """Essentially just Iterator[List[T]]. This class is just here so that I can generate the docs with nicely formatted types like "Table[str]".""" # Table
    pass                                                                         # Table
Table._name = "Table"                                                            # Table
#Table.__module__ = "cli"                                                        # Table
class Row(list):                                                                 # Row
    """Not really used currently. Just here for potential future feature"""      # Row
    pass                                                                         # Row
_jsFAuto = k1lib.AutoIncrement(prefix=f"_jsF_{random.randint(100, 999)}_")       # Row
_jsDAuto = k1lib.AutoIncrement(prefix=f"_jsD_{random.randint(100, 999)}_")       # Row
class BaseCli:                                                                   # BaseCli
    """A base class for all the cli stuff. You can definitely create new cli tools that
have the same feel without extending from this class, but advanced stream operations
(like ``+``, ``&``, ``.all()``, ``|``) won't work.

At the moment, you don't have to call super().__init__() and super().__ror__(),
as __init__'s only job right now is to solidify any :class:`~k1lib.cli.modifier.op`
passed to it, and __ror__ does nothing."""                                       # BaseCli
    def __init__(self, fs:list=[], capture=False):                               # BaseCli
        """Not expected to be instantiated by the end user.

**fs param**

Expected to use it like this::

    class A(BaseCli):
        def __init__(self, f):
            fs = [f]; super().__init__(fs); self.f = fs[0]

Where ``f`` is some (potentially exotic) function. This will replace f with a "normal"
function that's executable. See source code of :class:`~k1lib.cli.filt.filt` for an
example of why this is useful. Currently, it will:

- Replace with last recorded ``4 in op()``, if ``f`` is :data:`True`, because Python does
  not allow returning complex objects from __contains__ method
- Solidifies every :class:`~k1lib.cli.modifier.op`.

:param capture: whether to capture all clis to the right of it and make it accessible under capturedClis and capturedSerial properties""" # BaseCli
        if isinstance(fs, tuple): raise AttributeError("`fs` should not be a tuple. Use a list instead, so that new functions can be returned") # BaseCli
        _k1_init_l = []                                                          # BaseCli
        for _k1_init_f in fs: cli.op.solidify(_k1_init_f); _k1_init_l.append(_k1_init_f) # this is supposed to turn the exotic function into a normal function and leave normal functions alone. Purposefully don't do heavy optimizations here, cause we might want to poke around and change its internal representation # BaseCli
        fs.clear(); fs.extend(_k1_init_l);                                       # BaseCli
        self.capture = capture; self._capturedClis = []; self._capturedSerial = None # BaseCli
    @property                                                                    # BaseCli
    def capturedClis(self):                                                      # BaseCli
        if isinstance(self._capturedClis, list):                                 # BaseCli
            ans = []                                                             # BaseCli
            for e in self._capturedClis: ans.append(cli.op.solidify(e))          # BaseCli
            self._capturedClis = tuple(ans)                                      # BaseCli
        return self._capturedClis                                                # BaseCli
    @property                                                                    # BaseCli
    def capturedSerial(self):                                                    # BaseCli
        if not self.capture: return None                                         # BaseCli
        if self._capturedSerial is None: self._capturedSerial = serial(*self.capturedClis) # BaseCli
        return self._capturedSerial                                              # BaseCli
    def hint(self, _hint:"cli.typehint.tBase"):                                  # BaseCli
        """Specifies output type hint."""                                        # BaseCli
        self._hint = _hint; return self                                          # BaseCli
    @property                                                                    # BaseCli
    def hasHint(self): return "_hint" in self.__dict__ and self._hint is not None # BaseCli
    def _typehint(self, inp:"cli.typehint.tBase"=None) -> "cli.typehint.tBase": return cli.typehint.tAny() if "_hint" not in self.__dict__ else self._hint # BaseCli
    def __and__(self, cli:"BaseCli") -> "oneToMany":                             # BaseCli
        """Duplicates input stream to multiple joined clis.
Example::

    # returns [[5], [0, 1, 2, 3, 4]]
    range(5) | (shape() & iden()) | deref()

Kinda like :class:`~k1lib.cli.modifier.apply`. There're just multiple ways of doing
this. This I think, is more intuitive, and :class:`~k1lib.cli.modifier.apply` is more
for lambdas and columns mode. Performances are pretty much identical."""         # BaseCli
        if isinstance(self, oneToMany): return self._copy()._after(cli)          # BaseCli
        if isinstance(cli, oneToMany): return cli._copy()._before(self)          # BaseCli
        return oneToMany(self, cli)                                              # BaseCli
    def __add__(self, cli:"BaseCli") -> "mtmS":                                  # BaseCli
        """Parallel pass multiple streams to multiple clis.
Example::

    # returns [8, 15]
    [2, 3] | ((op() * 4) + (op() * 5)) | deref()"""                              # BaseCli
        if isinstance(self, mtmS): return self._copy()._after(cli)               # BaseCli
        if isinstance(cli, mtmS): return cli._copy()._before(self)               # BaseCli
        return mtmS(self, cli)                                                   # BaseCli
    def all(self, n:int=1) -> "BaseCli":                                         # BaseCli
        """Applies this cli to all incoming streams.
Example::

    # returns (3,)
    torch.randn(3, 4) | toMean().all() | shape()
    # returns (3, 4)
    torch.randn(3, 4, 5) | toMean().all(2) | shape()

:param n: how many times should I chain ``.all()``?"""                           # BaseCli
        if n < 0: raise AttributeError(f"Does not make sense for `n` to be \"{n}\"") # BaseCli
        s = self                                                                 # BaseCli
        for i in range(n): s = cli.apply(s)                                      # BaseCli
        return s                                                                 # BaseCli
    def __or__(self, cli_) -> "[cli, captureSuccessful]": # cli is guaranteed (by typical usage, not law) that it's a BaseCli # BaseCli
        """Joins clis end-to-end.
Example::

    c = apply(op() ** 2) | deref()
    # returns [0, 1, 4, 9, 16]
    range(5) | c"""                                                              # BaseCli
        if not isinstance(self, cli.op) and hasattr(self, "capture") and self.capture: self._capturedClis.append(cli_); return self # BaseCli
        if isinstance(self, serial): return self._after(cli_)                    # BaseCli
        if isinstance(cli_, serial): return cli_._before(self)                   # BaseCli
        return serial(self, cli_)                                                # BaseCli
    def __ror__(self, it): return NotImplemented                                 # BaseCli
    def f(self) -> Table[Table[int]]:                                            # BaseCli
        """Creates a normal function :math:`f(x)` which is equivalent to
``x | self``."""                                                                 # BaseCli
        return lambda it: self.__ror__(it)                                       # BaseCli
    def __lt__(self, it):                                                        # BaseCli
        """Backup pipe symbol `>`, purely for style, so that you can do something like
this::

    range(4) > file("a.txt")"""                                                  # BaseCli
        return self.__ror__(it)                                                  # BaseCli
    def __call__(self, it, *args):                                               # BaseCli
        """Another way to do ``it | cli``. If multiple arguments are fed, then the
argument list is passed to cli instead of just the first element. Example::

    @applyS
    def f(it):
        return it
    f(2) # returns 2
    f(2, 3) # returns [2, 3]"""                                                  # BaseCli
        if len(args) == 0: return self.__ror__(it)                               # BaseCli
        else: return self.__ror__([it, *args])                                   # BaseCli
    def __neg__(self):                                                           # BaseCli
        """Alias for __invert__, for clis that support inverting stuff."""       # BaseCli
        return ~self                                                             # BaseCli
    def _all_array_opt(self, it, level:int):                                     # BaseCli
        """Array types optimization for ``operator.all(level)``.

Essentially, a lot of times, I'm trying to do ``array | op()[3].all()``,
or ``array | transpose().all()``. But without this optimization, that ``.all()``
function kinda loops through each element and operates on them in vanilla Python,
which is super slow. So, this is a mechanism to speed it up. Here's how it works::

    # you wrote this
    array | operator.all() | deref()
    # apply() detects that you're trying to operate on an array type. It then figures
    # out how many nested apply() levels are there. In this case it's 1, so apply() returns this instead
    operator._all_array_opt(array, 1)
    # if that throws an error or returns NotImplemented, then it'll just loop through the array normally

    # if you wrote this instead
    array | operator.all(3) | deref()
    # or this
    array | apply(apply(operator.all())) | deref()
    # apply() will try to execute this instead
    operator._all_array_opt(array, 3)

Also, if the operator is a complex one, made of an entire pipeline, then ``serial`` can break
them apart and do this kind of optimization on each simple operator like this::

    operator = op()[3] | transpose()
    array | operator.all() | deref()
    # that gets transformed into this
    array | op()[3].all() | transpose().all() | deref()
    # then, array() will be called 2 times
    arr2 = op()[3]._all_array_opt(array, 1)
    transpose()._all_array_opt(arr2, 1)

It also works on something more complicated and nested like this::

    # returns
    np.random.randn(3,4,5,6,7,8) | apply(transpose().all(3) | item()) | shape()

This breakdown also happens with op() (anticipated feature, not implemented yet)::

    array | op()[3][:4].all() | deref()
    # this will be broken down into
    array | op()[3].all() | op()[:4].all() | deref()
    # each piece will now have a chance to optimize the array structure independently,
    # so even if op()[:4] can't be done, op()[3] still have a chance to do the C-optimized version

Why don't I build a more standardized structure for these optimization passes? Well
I did, along the lines of LLVM. But, the whole optimization process kinda takes a long
time and I'm not sure if it's truly flexible for the kinds of workloads that I'm thinking
about. So, I'll just do this quick dumb optimization hack to get it over with, and when
I can think more clearly about this, I might move this mechanism back to LLVM.""" # BaseCli
        return NotImplemented                                                    # BaseCli
    def _jsF(self, meta):                                                        # BaseCli
        """JS transpiler default function. See "JS transpiler" section in the docs""" # BaseCli
        return NotImplemented                                                    # BaseCli
    def _pyF(self, meta):                                                        # BaseCli
        """Cli to Python transpiler default function."""                         # BaseCli
        return NotImplemented                                                    # BaseCli
    def _cppF(self, meta):                                                       # BaseCli
        """C++ transpiler default function"""                                    # BaseCli
        return NotImplemented                                                    # BaseCli
    def _javaF(self, meta):                                                      # BaseCli
        """Java transpiler default function"""                                   # BaseCli
        return NotImplemented                                                    # BaseCli
    def _sqlF(self, meta):                                                       # BaseCli
        """SQL transpiler default function"""                                    # BaseCli
        return NotImplemented                                                    # BaseCli
def _k1_init_frames():                                                           # _k1_init_frames
    _k1_init_frames_count = 0                                                    # _k1_init_frames
    try:                                                                         # _k1_init_frames
        while True:                                                              # _k1_init_frames
            yield sys._getframe(_k1_init_frames_count) # `sys._getframe()` trick stolen from pd.DataFrame.query # _k1_init_frames
            _k1_init_frames_count += 1                                           # _k1_init_frames
    except: pass                                                                 # _k1_init_frames
def _k1_global_frame():                                                          # _k1_global_frame
    try:                                                                         # _k1_global_frame
        _k1_init_frames_ans = {}                                                 # _k1_global_frame
        for _k1_init_frames_frame in reversed(list(_k1_init_frames())):          # _k1_global_frame
            _k1_init_frames_ans = {**_k1_init_frames_ans, **_k1_init_frames_frame.f_locals} # _k1_global_frame
        return _k1_init_frames_ans                                               # _k1_global_frame
    except: return {}                                                            # _k1_global_frame
def fastF(c, x=None):                                                            # fastF
    """Tries to figure out what's going on, is it a normal function, or an applyS,
or a BaseCli, etc., and return a really fast function for execution. Example::

    # both returns 16, fastF returns "lambda x: x**2", so it's really fast
    fastF(op()**2)(4)
    fastF(applyS(lambda x: x**2))(4)

At the moment, parameter ``x`` does nothing, but potentially in the future, you can
pass in an example input to the cli, so that this returns an optimized, C compiled
version.

:param x: sample data for the cli"""                                             # fastF
    if isinstance(c, str): return fastF(eval(f"lambda x: {c}", _k1_global_frame())) # fastF
    if isinstance(c, cli.op): return c.ab_fastF()                                # fastF
    if isinstance(c, cli.applyS):                                                # fastF
        f = fastF(c.f)                                                           # fastF
        if len(c.args) == 0 and len(c.kwargs) == 0: return f                     # fastF
        else: return lambda x, *args, **kwargs: f(x, *c.args, **c.kwargs)        # fastF
    if isinstance(c, BaseCli): return c.__ror__                                  # fastF
    return c                                                                     # fastF
def checkRor(c):                                                                 # checkRor
    if isinstance(c, BaseCli): return c                                          # checkRor
    if hasattr(c, "__ror__"): return cli.aS(c.__ror__)                           # checkRor
    if callable(c): return cli.aS(c)                                             # checkRor
    raise Exception(f"Trying to add an operator to the pipeline, but the given object is not derived from BaseCli nor does it define a __ror__ method") # checkRor
class serialRepeat(BaseCli):                                                     # serialRepeat
    def __init__(self, f, n:int):                                                # serialRepeat
        self.f = f; self.n = n; self._fC = fastF(f)                              # serialRepeat
    def __ror__(self, it):                                                       # serialRepeat
        f = self._fC                                                             # serialRepeat
        for i in range(self.n): it = f(it)                                       # serialRepeat
        return it                                                                # serialRepeat
    def _jsF(self, meta):                                                        # serialRepeat
        f = self.f; fIdx = _jsFAuto(); dataIdx = _jsDAuto(); res = k1lib.kast.asyncGuard(k1lib.kast.prepareFunc3(f, meta)) # serialRepeat
        if res is NotImplemented: raise Exception(f"{cli.__class__} can't be transpiled into js. Either it doesn't make sense, or it hasn't been built yet") # serialRepeat
        header, fn, _async = res                                                 # serialRepeat
        return f"""\
{header}\nconst {fIdx} = {'async ' if _async else ''}({dataIdx}) => {{
    for (let i = 0; i < {self.n}; i++) {{
        {dataIdx} = {'await ' if _async else ''}{fn}({dataIdx});
    }}
    return {dataIdx};
}}""", fIdx                                                                      # serialRepeat
class serial(BaseCli):                                                           # serial
    def __init__(self, *clis:List[BaseCli]):                                     # serial
        """Merges clis into 1, feeding end to end. Used in chaining clis
together without a prime iterator. Meaning, without this, stuff like this
fails to run::

    [1, 2] | a() | b() # runs
    c = a() | b(); [1, 2] | c # doesn't run if this class doesn't exist"""       # serial
        fs = [checkRor(c) for c in clis]; super().__init__(fs); self.clis = fs; self._runOpt() # serial
    def _runOpt(self):                                                           # serial
        self._hasTrace = any(isinstance(c, cli.trace) for c in self.clis)        # serial
        self._cliCs = [fastF(c) for c in self.clis]; return self                 # serial
    def _typehint(self, inp=None):                                               # serial
        for c in self.clis: inp = c._typehint(inp) or cli.typehint.tAny()        # serial
        return inp                                                               # serial
    def __ror__(self, it:Iterator[Any]) -> Iterator[Any]:                        # serial
        if self._hasTrace: # slower, but tracable                                # serial
            for cli in self._clis_after_capture_analysis: it = it | cli          # serial
        else: # faster, but not tracable                                         # serial
            for cli in self._cliCs: it = cli(it)                                 # serial
        return it                                                                # serial
    def _before(self, c): return serial(checkRor(c), *self.clis)                 # serial
    def _after (self, c): return serial(*self.clis, checkRor(c))                 # serial
    def _jsF(self, meta):                                                        # serial
        headers = []; fns = []; asyncs = []; fIdx = _jsFAuto(); dataIdx = _jsDAuto() # serial
        for cli in self.clis:                                                    # serial
            res = k1lib.kast.asyncGuard(cli._jsF(meta))                          # serial
            if res is NotImplemented: raise Exception(f"{cli.__class__} can't be transpiled into js. Either it doesn't make sense, or it hasn't been built yet") # serial
            header, fn, _async = res                                             # serial
            headers.append(header); fns.append(fn); asyncs.append(_async)        # serial
        body = dataIdx                                                           # serial
        for fn, _async in zip(fns, asyncs): body = f"{'await ' if _async else ''}{fn}({body})" # serial
        return "\n".join(headers) + f"""\nconst {fIdx} = {'async ' if any(asyncs) else ''}({dataIdx}) => {{ return {body}; }};""", fIdx # serial
    @staticmethod                                                                # serial
    def repeat(f, n:int):                                                        # serial
        """Executes this function over and over again for n times.
Example::

    # returns 6561, or ((3^2)^2)^2
    3 | serial.repeat(op()**2, 3)

Of course, you can also do something like this::

    3 | serial(*[lambda x: x**2]*3)

And it would achieve the same result, but using this method, you can
vary n if you were to transpile it to JS.

:param f: function to execute
:param n: how many times to execute this function serially"""                    # serial
        return serialRepeat(f, n)                                                # serial
atomic.add("baseAnd", (Number, np.number, str, dict, bool, bytes, list, tuple, *([torch.Tensor] if hasTorch else []), np.ndarray, xml.etree.ElementTree.Element), "used by BaseCli.__and__") # serial
def addAtomic(klass):                                                            # addAtomic
    atomic.baseAnd = (*atomic.baseAnd, klass)                                    # addAtomic
    atomic.deref = (*atomic.deref, klass)                                        # addAtomic
def _iterable(it):                                                               # _iterable
    try: iter(it); return True                                                   # _iterable
    except: return False                                                         # _iterable
class oneToMany(BaseCli):                                                        # oneToMany
    def __init__(self, *clis:List[BaseCli]):                                     # oneToMany
        """Duplicates 1 stream into multiple streams, each for a cli in the
list. Used in the "a & b" joining operator. See also: :meth:`BaseCli.__and__`""" # oneToMany
        fs = [checkRor(c) for c in clis]; super().__init__(fs); self.clis = fs; self._cache() # oneToMany
    def _typehint(self, inp):                                                    # oneToMany
        ts = []                                                                  # oneToMany
        for f in self.clis:                                                      # oneToMany
            try: ts.append(f._typehint(inp))                                     # oneToMany
            except: ts.append(cli.typehint.tAny())                               # oneToMany
        return cli.typehint.tCollection(*ts).reduce()                            # oneToMany
    def __ror__(self, it:Iterator[Any]) -> Iterator[Iterator[Any]]:              # oneToMany
        if isinstance(it, atomic.baseAnd) or isinstance(it, k1lib.cli.splitSeek) or not _iterable(it): # oneToMany
            for cli in self._cliCs: yield cli(it)                                # oneToMany
        else:                                                                    # oneToMany
            its = itertools.tee(it, len(self.clis))                              # oneToMany
            for cli, it in zip(self._cliCs, its): yield cli(it)                  # oneToMany
    def _cache(self): self._cliCs = [fastF(c) for c in self.clis]; return self   # oneToMany
    def _before(self, c): self.clis = [checkRor(c)] + self.clis; return self._cache() # oneToMany
    def _after(self, c): self.clis = self.clis + [checkRor(c)]; return self._cache() # oneToMany
    def _copy(self): return oneToMany(*self.clis)                                # oneToMany
    def _jsF(self, meta):                                                        # oneToMany
        headers = []; fns = []; asyncs = []; fIdx = _jsFAuto(); dataIdx = _jsDAuto() # oneToMany
        for cli in self.clis:                                                    # oneToMany
            res = k1lib.kast.asyncGuard(cli._jsF(meta))                          # oneToMany
            if res is NotImplemented: raise Exception(f"{cli.__class__} can't be transpiled into js. Either it doesn't make sense, or it hasn't been built yet") # oneToMany
            header, fn, _async = res                                             # oneToMany
            headers.append(header); fns.append(fn); asyncs.append(_async)        # oneToMany
        body = "[" + ", ".join([f"{'await ' if _async else ''}{fn}({dataIdx})" for fn, _async in zip(fns, asyncs)]) + "]" # oneToMany
        return "\n".join(headers) + f"""\nconst {fIdx} = {'async ' if any(asyncs) else ''}({dataIdx}) => {body};""", fIdx # oneToMany
class mtmS(BaseCli):                                                             # mtmS
    def __init__(self, *clis:List[BaseCli]):                                     # mtmS
        """Applies multiple streams to multiple clis independently. Used in
the "a + b" joining operator. See also: :meth:`BaseCli.__add__`.

Weird name is actually a shorthand for "many to many specific"."""               # mtmS
        fs = [checkRor(c) for c in clis]; super().__init__(fs=fs); self.clis = fs; self._cache() # mtmS
    def _inpTypeHintExpand(self, t):                                             # mtmS
        n = len(self.clis);                                                      # mtmS
        if isinstance(t, (cli.typehint.tCollection, *cli.typehint.tListIterSet, cli.typehint.tArrayTypes)): return t.expand(n) # mtmS
        else: return [cli.typehint.tAny()]*n                                     # mtmS
    def _typehint(self, t):                                                      # mtmS
        n = len(self.clis); outTs = []                                           # mtmS
        for c, t in zip(self.clis, self._inpTypeHintExpand(t)):                  # mtmS
            try: outTs.append(c._typehint(t))                                    # mtmS
            except: outTs.append(cli.typehint.tAny())                            # mtmS
        return cli.typehint.tCollection(*outTs).reduce()                         # mtmS
    def _cache(self): self._cliCs = [fastF(c) for c in self.clis]; return self   # mtmS
    def _before(self, c): self.clis = [checkRor(c)] + self.clis; return self._cache() # mtmS
    def _after (self, c): self.clis = self.clis + [checkRor(c)]; return self._cache() # mtmS
    def __ror__(self, its:Iterator[Any]) -> Iterator[Any]:                       # mtmS
        for cli, it in zip(self._cliCs, its): yield cli(it)                      # mtmS
    @staticmethod                                                                # mtmS
    def f(f, i:int, n:int=100):                                                  # mtmS
        """Convenience method, so
that this::

    mtmS(iden(), op()**2, iden(), iden(), iden())
    # also the same as this btw:
    (iden() + op()**2 + iden() + iden() + iden())

is the same as this::

    mtmS.f(op()**2, 1, 5)

Example::

    # returns [5, 36, 7, 8, 9]
    range(5, 10) | mtmS.f(op()**2, 1, 5) | deref()

:param i: where should I put the function?
:param n: how many clis in total? Defaulted to 100"""                            # mtmS
        return mtmS(*([cli.iden()]*i + [f] + [cli.iden()]*(n-i-1)))              # mtmS
    def _copy(self): return mtmS(*self.clis)                                     # mtmS
    def _jsF(self, meta):                                                        # mtmS
        headers = []; fns = []; asyncs = []; fIdx = _jsFAuto(); dataIdx = _jsDAuto() # mtmS
        for cli in self.clis:                                                    # mtmS
            res = k1lib.kast.asyncGuard(cli._jsF(meta))                          # mtmS
            if res is NotImplemented: raise Exception(f"{cli.__class__} can't be transpiled into js. Either it doesn't make sense, or it hasn't been built yet") # mtmS
            header, fn, _async = res                                             # mtmS
            headers.append(header); fns.append(fn); asyncs.append(_async)        # mtmS
        body = "[" + ", ".join([f"{'await ' if _async else ''}{fn}({dataIdx}[{i}])" for i, (fn, _async) in enumerate(zip(fns, asyncs))]) + "]" # mtmS
        return "\n".join(headers) + f"""\nconst {fIdx} = {'async ' if any(asyncs) else ''}({dataIdx}) => {body};""", fIdx # mtmS
def patchNumpy():                                                                # patchNumpy
    """Patches numpy arrays and data types, so that piping like
this work::

    a = np.random.randn(3)
    a | shape() # returns (3,)"""                                                # patchNumpy
    try:                                                                         # patchNumpy
        if np._k1_patched: return                                                # patchNumpy
    except: pass                                                                 # patchNumpy
    try:                                                                         # patchNumpy
        import forbiddenfruit, inspect; #forbiddenfruit.reverse(np.ndarray, "__or__") # old version # patchNumpy
        oldOr = np.ndarray.__or__                                                # patchNumpy
        def _newNpOr(self, v):                                                   # patchNumpy
            if isinstance(v, BaseCli): return NotImplemented                     # patchNumpy
            try: return oldOr(self, v)                                           # patchNumpy
            except: warnings.warn(traceback.format_exc())                        # patchNumpy
        forbiddenfruit.curse(np.ndarray, "__or__", _newNpOr)                     # patchNumpy
        a = [getattr(np, dk) for dk in np.__dict__.keys()] # patching all numpy's numeric types # patchNumpy
        for _type in [x for x in a if inspect.isclass(x) and issubclass(x, np.number) and not issubclass(x, np.integer)]: # patchNumpy
            _oldOr = _type.__or__                                                # patchNumpy
            def _typeNewOr(self, v):                                             # patchNumpy
                if isinstance(v, BaseCli): return NotImplemented                 # patchNumpy
                try: return _oldOr(self, v)                                      # patchNumpy
                except: warnings.warn(traceback.format_exc())                    # patchNumpy
            forbiddenfruit.curse(_type, "__or__", _typeNewOr)                    # patchNumpy
        np._k1_patched = True                                                    # patchNumpy
    except Exception as e: warnings.warn(f"Tried to patch __or__ operator of built-in type `np.ndarray` but can't because: {e}") # patchNumpy
dict_keys = type({"a": 3}.keys());   oldDKOr = dict_keys.__or__                  # patchNumpy
dict_items = type({"a": 3}.items()); oldDIOr = dict_items.__or__                 # patchNumpy
oldSetOr = set.__or__                                                            # patchNumpy
def patchDict():                                                                 # patchDict
    """Patches dictionaries's items and keys, so that piping
works::

    d = {"a": 3, "b": 4}
    d.keys() | deref() # returns ["a", "b"]
    d.items() | deref() # returns [["a", 3], ["b", 4]]"""                        # patchDict
    try:                                                                         # patchDict
        if np._k1_dict_patched: return                                           # patchDict
    except: pass                                                                 # patchDict
    try:                                                                         # patchDict
        import forbiddenfruit, traceback                                         # patchDict
        def _newDOr(self, v):                                                    # patchDict
            """Why is this so weird? For some reason, if you patch dict_keys, you will
            also patch dict_items. So, if you were to have 2 functions, one for each,
            then they will override each other. The way forward is to have 1 single
            function detect whether it's dict_keys or dict_items, and call the correct
            original function. So why are there 2 curses? Well cause I'm lazy to check
            for this behavior in multiple python versions, so just have 2 to make sure.""" # patchDict
            if isinstance(v, BaseCli): return NotImplemented                     # patchDict
            try:                                                                 # patchDict
                # print(self, type(self), v, type(v))                            # patchDict
                if isinstance(self, dict_keys): return oldDKOr(self, v)          # patchDict
                elif isinstance(self, dict):                                     # patchDict
                    if isinstance(v, dict_keys): return oldSetOr(set(self.keys()), set(v)) # patchDict
                    return oldDIOr(self, v)                                      # patchDict
                elif isinstance(self, set):                                      # patchDict
                    if isinstance(v, dict_keys): return oldSetOr(self, set(v))   # patchDict
                    return oldSetOr(self, v)                                     # patchDict
            except:                                                              # patchDict
                print(self, type(self), v, type(v))                              # patchDict
                warnings.warn(traceback.format_exc())                            # patchDict
                return NotImplemented                                            # patchDict
        forbiddenfruit.curse(dict_keys, "__or__", _newDOr)                       # patchDict
        forbiddenfruit.curse(dict_items, "__or__", _newDOr)                      # patchDict
        np._k1_dict_patched = True                                               # patchDict
    except Exception as e: warnings.warn(f"Tried to patch __or__ operator of built-in type `dict_keys` and `dict_items` but can't because: {e}") # patchDict
def patchPandas():                                                               # patchPandas
    """Patches panda's :class:`pandas.core.series.Series` and
:class:`pandas.core.frame.DataFrame` so that piping works::

    pd.read_csv("a.csv")["col3"] | shape()"""                                    # patchPandas
    try:                                                                         # patchPandas
        import pandas as pd                                                      # patchPandas
    except: return                                                               # patchPandas
    try:                                                                         # patchPandas
        if pd._k1_patched: return                                                # patchPandas
    except: pass                                                                 # patchPandas
    try:                                                                         # patchPandas
        import forbiddenfruit                                                    # patchPandas
        oldPdSOr = pd.core.series.Series.__or__                                  # patchPandas
        def _newPdSOr(self, v):                                                  # patchPandas
            if isinstance(v, BaseCli): return NotImplemented                     # patchPandas
            try: return oldPdSOr(self, v)                                        # patchPandas
            except: warnings.warn(traceback.format_exc())                        # patchPandas
        forbiddenfruit.curse(pd.core.series.Series, "__or__", _newPdSOr)         # patchPandas
                                                                                 # patchPandas
        oldPdDFOr = pd.core.frame.DataFrame                                      # patchPandas
        def _newPdDFOr(self, v):                                                 # patchPandas
            if isinstance(v, BaseCli): return NotImplemented                     # patchPandas
            try: return oldPdDFOr(self, v)                                       # patchPandas
            except: warnings.warn(traceback.format_exc())                        # patchPandas
        forbiddenfruit.curse(pd.core.frame.DataFrame, "__or__", _newPdDFOr)      # patchPandas
        pd._k1_patched = True                                                    # patchPandas
    except Exception as e: warnings.warn(f"Tried to patch __or__ operator of built-in type `pd.core.series.Series` but can't because: {e}") # patchPandas



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
import k1lib, time; from k1lib import fmt
from .callbacks import Callback, Callbacks, Cbs
import k1lib.cli as cli; plt = k1lib.dep("matplotlib.pyplot")
__all__ = ["ProgressBar"]
@k1lib.patch(Cbs)
class ProgressBar(Callback):                                                     # ProgressBar
    """Displays the current progress, epoch and batch while running.
Deposits variables into :class:`~k1lib.Learner` at checkpoint ``startBatch``:

- progress: single float from 0 to 1, guaranteed to increase monotonically
- epochThroughput
- remaining: estimated remaining time. Does not take into account callbacks that can potentially cancel the run prematurely""" # ProgressBar
    def startRun(self):                                                          # ProgressBar
        self.startTime = time.time(); self.step = 0; self.l.progress = 0         # ProgressBar
        self.l.loss = float("inf") # to make sure this variable exist            # ProgressBar
        self.aL = self.bL = self.cL = self.dL = self.eL = 0; self.data = []      # ProgressBar
    def startBatch(self):                                                        # ProgressBar
        batch = self.l.batch; batches = self.l.batches; epoch = self.l.epoch; epochs = self.l.epochs # ProgressBar
        elapsedTime = self.elapsedTime = time.time() - self.startTime            # ProgressBar
        if batches is None: progress = self.l.progress = epoch / epochs; batchTh = None # ProgressBar
        else:                                                                    # ProgressBar
            progress = self.l.progress = (batch / batches + epoch) / epochs      # ProgressBar
            batchTh = batches * epochs * progress / elapsedTime                  # ProgressBar
        epochTh = self.l.epochThroughput = epochs * progress / elapsedTime; self.l.batchThroughput = batchTh # ProgressBar
        remaining = self.l.remaining = round(elapsedTime / (progress+1e-7) * (1-progress), 2) if progress > 0 else float('inf') # ProgressBar
                                                                                 # ProgressBar
        a = str(round(100 * progress)); self.aL = max(self.aL, len(a)); a = a.rjust(self.aL) # ProgressBar
        b = f"{epoch}/{epochs} ({fmt.throughput(epochTh, ' epochs')})"; self.bL = max(self.bL, len(b)); b = b.rjust(self.bL) # ProgressBar
        if batches is not None:                                                  # ProgressBar
            c = f"{batch}/{batches} ({fmt.throughput(batchTh, ' batches')})"; self.cL = max(self.cL, len(c)); c = c.rjust(self.cL) # ProgressBar
        else: c = f"{batch}/{batches}"; self.cL = max(self.cL, len(c)); c = c.rjust(self.cL) # ProgressBar
        d = f"{round(elapsedTime, 2)}".rjust(6); self.dL = max(self.dL, len(d)); d = d.rjust(self.dL) # ProgressBar
        e = f"{remaining}"; self.eL = max(self.eL, len(e)); e = e.rjust(self.eL) # ProgressBar
        self.data.append([epoch, batch, elapsedTime, progress, epochTh, batchTh, remaining]) # ProgressBar
        print(f"\rProgress: {a}%, epoch: {b}, batch: {c}, elapsed: {d}s, remaining: {e}s, loss: {self.l.loss}             ", end="") # ProgressBar
    def plot(self, f=cli.iden(), perEpoch=False, _window=2):                     # ProgressBar
        """Plots detailed partial execution time profile.

:param f: optional post processing step
:param perEpoch: if True, normalize time per epoch, else keep it at time per run
:param _window: number of batches to calculate the processing rate over. Put low
    values (min 2) to make it crisp (and inaccurate), put high values to make it
    smooth (and accurate)"""                                                     # ProgressBar
        if perEpoch: f = cli.apply(cli.op()/self.l.epochs) | f                   # ProgressBar
        self.data | cli.cut(2, 3) | cli.deref() | cli.window(_window)\
        | cli.apply(cli.rows(0, -1) | cli.transpose() | ~cli.apply(lambda x, y: y-x) | ~cli.aS(lambda x, y: x/y))\
        | f | cli.deref() | cli.aS(plt.plot)                                     # ProgressBar
        plt.ylabel("Time/epoch (seconds)" if perEpoch else "Time/run (seconds)"); # ProgressBar



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""Higher order functions"""
__all__ = ["Func", "polyfit", "derivative", "optimize", "inverse", "integral", "batchify"]
from typing import Callable, List
import k1lib, numpy as np, warnings, threading, time, inspect
from functools import partial
plt = k1lib.dep("matplotlib.pyplot")
import k1lib.cli as cli
Func = Callable[[float], float]
def polyfit(x:List[float], y:List[float], deg:int=6) -> Func:                    # polyfit
    """Returns a function that approximate :math:`f(x) = y`.
Example::

    xs = [1, 2, 3]
    ys = [2, 3, 5]
    f = k1.polyfit(xs, ys, 1)

This will create a best-fit function. You can just use it as a regular,
normal function. You can even pass in :class:`numpy.ndarray`::

    # returns some float
    f(2)

    # plots fit function from 0 to 5
    xs = np.linspace(0, 5)
    plt.plot(xs, f(xs))

:param deg: degree of the polynomial of the returned function"""                 # polyfit
    params = np.polyfit(x, y, deg)                                               # polyfit
    def _inner(_x):                                                              # polyfit
        answer = np.zeros_like(_x, dtype=float)                                  # polyfit
        for expo, param in enumerate(params):                                    # polyfit
            answer += param * _x**(len(params)-expo-1)                           # polyfit
        return answer                                                            # polyfit
    return _inner                                                                # polyfit
def derivative(f:Func, delta:float=1e-6) -> Func:                                # derivative
    """Returns the derivative of a function.
Example::

    f = lambda x: x**2
    df = k1lib.derivative(f)
    df(3) # returns roughly 6 """                                                # derivative
    return lambda x: (f(x + delta) - f(x)) / delta                               # derivative
def optimize(f:Func, v:float=1, threshold:float=1e-6, **kwargs) -> float:        # optimize
    r"""Given :math:`f(x) = 0`, solves for x using Newton's method with initial value
`v`. Example::

    f = lambda x: x**2-2
    # returns 1.4142 (root 2)
    k1lib.optimize(f)
    # returns -1.4142 (negative root 2)
    k1lib.optimize(f, -1)

Interestingly, for some reason, result of this is more accurate than :meth:`derivative`.
"""                                                                              # optimize
    if len(kwargs) > 0: f = partial(f, **kwargs)                                 # optimize
    fD = derivative(f)                                                           # optimize
    for i in range(20):                                                          # optimize
        v = v - f(v)/fD(v)                                                       # optimize
    if abs(f(v)) > threshold: warnings.warn("k1lib.optimize not converging")     # optimize
    return v                                                                     # optimize
def inverse(f:Func) -> Func:                                                     # inverse
    """Returns the inverse of a function.
Example::

    f = lambda x: x**2
    fInv = k1lib.inverse(f)
    # returns roughly 3
    fInv(9)

.. warning::
    The inverse function takes a long time to run, so don't use this
    where you need lots of speed. Also, as you might imagine, the
    inverse function isn't really airtight. Should work well with
    monotonic functions, but all bets are off with other functions."""           # inverse
    return lambda y: optimize(lambda x: f(x) - y)                                # inverse
def integral(f:Func, _range:k1lib.Range) -> float:                               # integral
    """Integrates a function over a range.
Example::

    f = lambda x: x**2
    # returns roughly 9
    k1lib.integral(f, [0, 3])

There is also the cli :class:`~k1lib.cli.modifier.integrate`
which has a slightly different api."""                                           # integral
    _range = k1lib.Range(_range)                                                 # integral
    n = 1000; xs = np.linspace(*_range, n)                                       # integral
    return sum([f(x)*_range.delta/n for x in xs])                                # integral
class Controller:                                                                # Controller
    def __init__(self):                                                          # Controller
        self.lock = threading.Lock(); self.data = {}; self.count = 0; self.event = threading.Event() # Controller
    def add(self, d):                                                            # Controller
        with self.lock: c = self.count; self.data[self.count] = d; self.count += 1; return c, self.event # Controller
    def prepare(self):                                                           # Controller
        with self.lock:                                                          # Controller
            data = {**self.data}; event = self.event                             # Controller
            self.data = {}; self.count = 0; self.event = threading.Event()       # Controller
            return data, event                                                   # Controller
def batchify(period=0.1) -> "singleFn":                                          # batchify
    """Transforms a function taking in batches to taking in singles,
for performance reasons.

Say you have this function that does some heavy computation::

    def f1(x, y):
        time.sleep(1)     # simulating heavy load, like loading large libraries/binaries
        return x + y      # does not take lots of time

Let's also say that a lot of time, you want to execute that function over multiple samples::

    res = [] # will be filled with [3, 7, 11]
    for x, y in [[1, 2], [3, 4], [5, 6]]:
        res.append(f1(x, y))

This would take 3 seconds to complete. But a lot of time, it might be advantageous to merge
them together and execute everything all at once::

    def f2(xs, ys):
        res = []; time.sleep(1)                       # loading large libraries/binaries only once
        for x, y in zip(xs, ys): res.append(x+y)      # run through all samples quickly
        return res

    res = f2([1, 3, 5], [2, 4, 6]) # filled with [3, 7, 11], just like before

But, may be you're in a multithreaded application and desire the original function "f1(x, y)",
instead of the batched function "f2(xs, ys)", like running an LLM (large language model) on
requests submitted by people on the internet. Each request that comes in runs on different
threads, but it's still desirable to pool together all of those requests, run through the model
once, and then split up the results to each respective request. That's where this functionality
comes in::

    @k1.batchify(0.1)     # pool up all calls every 0.1 seconds
    def f2(xs, ys):
        ...               # same as previous example
    @k1.batchify          # can also do it like this. It'll default to a period of 0.1
    def f2(xs, ys):
        ...               # same as previous example

    res = []
    t1 = threading.Thread(target=lambda: res.append(f(1, 2)))
    t2 = threading.Thread(target=lambda: res.append(f(3, 4)))
    t3 = threading.Thread(target=lambda: res.append(f(5, 6)))
    ths = [t1, t2, t3]
    for th in ths: th.start()
    for th in ths: th.join() # after this point, res will have a permutation of [3, 7, 11] (because t1, t2, t3 execution order is not known)

This will take on average 1 + 0.1 seconds (heavy load execution time + refresh rate). You can
decorate this on any Flask endpoint you want and it will just work::

    @app.get("/run/<int:nums>")
    @k1.batchify(0.3)
    def run(nums):
        time.sleep(1) # long running process
        return nums | apply("x**2") | apply(str) | deref()

Now, if you send 10 requests within a window of 0.3s, then the total running time would only be
1.3s, instead of 10s like before"""                                              # batchify
    def wrap(batchedFn):                                                         # batchify
        data = []; con = Controller()                                            # batchify
        def bgProcess():                                                         # batchify
            while True:                                                          # batchify
                time.sleep(period); data, event = con.prepare()                  # batchify
                if len(data) > 0:                                                # batchify
                    d = data.values(); sentinel = object()                       # batchify
                    args = d | cli.cut(1) | cli.transpose(fill=sentinel) | cli.aS(lambda x: None if x is sentinel else x).all(2) | cli.deref() # batchify
                    kws = d | cli.cut(2) | cli.deref()                           # batchify
                    keys = kws | cli.op().keys().all() | cli.joinSt() | cli.aS(set) | cli.deref(); keys # batchify
                    kws = keys | cli.apply(lambda key: [key, kws | cli.apply(lambda kw: kw.get(key, None))]) | cli.toDict() | cli.deref() # batchify
                                                                                 # batchify
                    res = list(batchedFn(*args, **kws))                          # batchify
                    for k, out in zip(data.keys(), res): data[k][0][0] = out     # batchify
                    event.set()                                                  # batchify
        threading.Thread(target=bgProcess).start()                               # batchify
        def inner(*args, **kwargs):                                              # batchify
            output = [...]; idx, event = con.add([output, args, kwargs])         # batchify
            t = threading.Thread(target=lambda: event.wait()); t.start(); t.join() # batchify
            return output[0]                                                     # batchify
        inner.fullargspec = inspect.getfullargspec(batchedFn)                    # batchify
        return inner                                                             # batchify
    if isinstance(period, (int, float)): return wrap                             # batchify
    else: batchedFn = period; period = 0.1; return wrap(batchedFn)               # batchify
